# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for# the specific language governing permissions and limitations
#  under the License.

import mock
import netaddr

import contextlib

from neutron.common import exceptions as q_exc

from quark.db import api as db_api
import quark.plugin_modules.mac_address_ranges as macrng_api
import quark.plugin_modules.networks as network_api
import quark.plugin_modules.ports as port_api
import quark.plugin_modules.subnets as subnet_api
from quark.tests.functional.base import BaseFunctionalTest


class QuarkFindPortsSorted(BaseFunctionalTest):
    def test_ports_sorted_by_created_at(self):
        # create a network
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        net_mod = db_api.network_create(self.context, **network)
        # create ports
        port1 = dict(network_id=net_mod["id"], backend_key="1", device_id="1")
        port2 = dict(network_id=net_mod["id"], backend_key="1", device_id="1")
        port3 = dict(network_id=net_mod["id"], backend_key="1", device_id="1")
        port_mod1 = db_api.port_create(self.context, **port1)
        port_mod2 = db_api.port_create(self.context, **port2)
        port_mod3 = db_api.port_create(self.context, **port3)
        res = db_api.port_find(self.context, None, None, None, None,
                               scope=db_api.ALL)
        self.assertTrue(res[0]["created_at"] < res[1]["created_at"] <
                        res[2]['created_at'])
        db_api.network_delete(self.context, net_mod)
        db_api.port_delete(self.context, port_mod1)
        db_api.port_delete(self.context, port_mod2)
        db_api.port_delete(self.context, port_mod3)


class QuarkCreatePortSatisfyIpam(BaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network_info, subnet_v4_info, subnet_v6_info=None):
        with contextlib.nested(
                mock.patch("neutron.common.rpc.get_notifier"),
                mock.patch("neutron.quota.QUOTAS.limit_check")):
            self.context.is_admin = True
            net = network_api.create_network(self.context, network_info)
            mac = {'mac_address_range': dict(cidr="AA:BB:CC")}
            macrng_api.create_mac_address_range(self.context, mac)
            self.context.is_admin = False
            subnet_v4_info['subnet']['network_id'] = net['id']
            sub_v4 = subnet_api.create_subnet(self.context, subnet_v4_info)
            sub_v6 = None
            if subnet_v6_info:
                subnet_v6_info['subnet']['network_id'] = net['id']
                sub_v6 = subnet_api.create_subnet(self.context, subnet_v6_info)

            yield net, sub_v4, sub_v6

    def test_port_created_should_satisfy_ipam_strategy(self):
        cidr = "192.168.1.0/24"
        cidr_v6 = "2001:db8::/32"
        ip_network = netaddr.IPNetwork(cidr)
        ipv6_network = netaddr.IPNetwork(cidr_v6)
        network = dict(id='1', name="public", tenant_id="make",
                       network_plugin="BASE",
                       ipam_strategy="BOTH_REQUIRED")
        network = {"network": network}
        subnet_v4 = dict(id='1', ip_version=4, next_auto_assign_ip=2,
                         cidr=cidr, first_ip=ip_network.first,
                         last_ip=ip_network.last, ip_policy=None,
                         tenant_id="fake")
        subnet_v6 = dict(id='2', ip_version=6, next_auto_assign_ip=2,
                         cidr=cidr_v6, first_ip=ipv6_network.first,
                         last_ip=ipv6_network.last, ip_policy=None,
                         tenant_id="fake")
        subnet_v4_info = {"subnet": subnet_v4}
        subnet_v6_info = {"subnet": subnet_v6}

        def _make_body(ipv4, ipv6):
            fix_ipv4 = dict(ip_address=ipv4, subnet_id=sub_v4['id'])
            fix_ipv6 = dict(ip_address=ipv6, subnet_id=sub_v6['id'])
            port_info = {"port": dict(fixed_ips=[fix_ipv4, fix_ipv6],
                                      network_id=net['id'])}
            return port_info

        with self._stubs(network, subnet_v4_info, subnet_v6_info) as (
                net, sub_v4, sub_v6):
            ipv4 = "192.168.1.50"
            ipv6 = "2001:db8::10"

            port = port_api.create_port(self.context, _make_body(ipv4, ipv6))
            port_ips = [ip["ip_address"] for ip in port["fixed_ips"]]

            for ip in [ipv4, ipv6]:
                self.assertTrue(ip in port_ips)

    def test_port_created_doesnt_satisfy_ipam_strategy_raises(self):
        cidr = "192.168.1.0/24"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(id='1', name="public", tenant_id="make",
                       network_plugin="BASE",
                       ipam_strategy="BOTH_REQUIRED")
        network = {"network": network}
        subnet_v4 = dict(id='1', ip_version=4, next_auto_assign_ip=2,
                         cidr=cidr, first_ip=ip_network.first,
                         last_ip=ip_network.last, ip_policy=None,
                         tenant_id="fake")
        subnet_v4_info = {"subnet": subnet_v4}

        def _make_body_only_v4(ip):
            fix_ip = dict(ip_address=ip, subnet_id=sub_v4['id'])
            port_info = {"port": dict(fixed_ips=[fix_ip],
                                      network_id=net['id'])}
            return port_info

        with self._stubs(network, subnet_v4_info) as (
                net, sub_v4, sub_v6):
            ip = "192.168.1.50"
            with self.assertRaises(q_exc.IpAddressGenerationFailure):
                port_api.create_port(self.context, _make_body_only_v4(ip))


class QuarkCreatePortWithIpNotMandatory(BaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network_info, subnet_v4_infos, subnet_v6_info=None):
        with contextlib.nested(
                mock.patch("neutron.common.rpc.get_notifier"),
                mock.patch("neutron.quota.QUOTAS.limit_check")):
            self.context.is_admin = True
            net = network_api.create_network(self.context, network_info)
            mac = {'mac_address_range': dict(cidr="AA:BB:CC")}
            macrng_api.create_mac_address_range(self.context, mac)
            self.context.is_admin = False
            sub_v4s = []
            for sub_info in subnet_v4_infos:
                sub_info['subnet']['network_id'] = net['id']
                sub_v4s.append(subnet_api.create_subnet(self.context,
                                                        sub_info))
            sub_v6 = None
            if subnet_v6_info:
                subnet_v6_info['subnet']['network_id'] = net['id']
                sub_v6 = subnet_api.create_subnet(self.context, subnet_v6_info)

            yield net, sub_v4s, sub_v6

    def test_port_created_with_only_subnet(self):
        cidr = "192.168.1.0/24"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(id='1', name="public", tenant_id="make",
                       network_plugin="BASE",
                       ipam_strategy="ANY")
        network = {"network": network}
        subnet_v4 = dict(id='1', ip_version=4, next_auto_assign_ip=2,
                         cidr=cidr, first_ip=ip_network.first,
                         last_ip=ip_network.last, ip_policy=None,
                         tenant_id="fake")
        subnet_v4_info = {"subnet": subnet_v4}

        def _make_body():
            fix_ipv4 = dict(subnet_id=sub_v4s[0]['id'])
            port_info = {"port": dict(fixed_ips=[fix_ipv4],
                                      network_id=net['id'])}
            return port_info

        with self._stubs(network, [subnet_v4_info]) as (
                net, sub_v4s, sub_v6):

            port = port_api.create_port(self.context, _make_body())
            port_ip = port["fixed_ips"][0]["ip_address"]

            self.assertTrue(netaddr.IPAddress(port_ip) in
                            netaddr.IPNetwork(cidr))

    def test_port_created_with_multiple_fixed_ips_some_only_with_subnet(self):
        cidr = "192.168.1.0/24"
        another_cidr = "192.168.2.0/24"
        cidr_v6 = "2001:db8::/32"
        ip_network = netaddr.IPNetwork(cidr)
        another_ip_network = netaddr.IPNetwork(another_cidr)
        ipv6_network = netaddr.IPNetwork(cidr_v6)
        network = dict(id='1', name="public", tenant_id="make",
                       network_plugin="BASE",
                       ipam_strategy="BOTH_REQUIRED")
        network = {"network": network}
        subnet_v4 = dict(id='1', ip_version=4, next_auto_assign_ip=2,
                         cidr=cidr, first_ip=ip_network.first,
                         last_ip=ip_network.last, ip_policy=None,
                         tenant_id="fake")
        another_subnet_v4 = dict(id='2', ip_version=4, next_auto_assign_ip=2,
                                 cidr=another_cidr,
                                 first_ip=another_ip_network.first,
                                 last_ip=another_ip_network.last,
                                 ip_policy=None, tenant_id="fake")
        subnet_v6 = dict(id='3', ip_version=6, next_auto_assign_ip=2,
                         cidr=cidr_v6, first_ip=ipv6_network.first,
                         last_ip=ipv6_network.last, ip_policy=None,
                         tenant_id="fake")
        subnet_v4_info = {"subnet": subnet_v4}
        another_subnet_v4_info = {"subnet": another_subnet_v4}
        subnet_v6_info = {"subnet": subnet_v6}

        def _make_body(ipv4, ipv6):
            fix_ipv4 = dict(ip_address=ipv4, subnet_id=sub_v4s[0]['id'])
            another_fix_ipv4 = dict(subnet_id=sub_v4s[1]['id'])
            fix_ipv6 = dict(ip_address=ipv6, subnet_id=sub_v6['id'])
            port_info = {"port": dict(fixed_ips=[fix_ipv4, another_fix_ipv4,
                                                 fix_ipv6],
                                      network_id=net['id'])}
            return port_info

        with self._stubs(network, [subnet_v4_info, another_subnet_v4_info],
                         subnet_v6_info) as (net, sub_v4s, sub_v6):
            ipv4 = "192.168.1.50"
            another_ipv4 = "192.168.2.1"
            ipv6 = "2001:db8::10"

            port = port_api.create_port(self.context, _make_body(ipv4, ipv6))
            port_ips = [ip["ip_address"] for ip in port["fixed_ips"]]

            for ip in [ipv4, another_ipv4, ipv6]:
                self.assertTrue(ip in port_ips)


class QuarkCreatePortWithForbiddenMacRange(BaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network_info, subnet_v4_infos, subnet_v6_info=None):
        with contextlib.nested(
                mock.patch("neutron.common.rpc.get_notifier"),
                mock.patch("neutron.quota.QUOTAS.limit_check")):
            self.context.is_admin = True
            net = network_api.create_network(self.context, network_info)
            mac = {'mac_address_range': dict(cidr="AA:BB:CC", do_not_use=True)}
            macrng_api.create_mac_address_range(self.context, mac)
            self.context.is_admin = False
            sub_v4s = []
            for sub_info in subnet_v4_infos:
                sub_info['subnet']['network_id'] = net['id']
                sub_v4s.append(subnet_api.create_subnet(self.context,
                                                        sub_info))
            sub_v6 = None
            if subnet_v6_info:
                subnet_v6_info['subnet']['network_id'] = net['id']
                sub_v6 = subnet_api.create_subnet(self.context, subnet_v6_info)

            yield net, sub_v4s, sub_v6

    def test_port_created_with_forbidden_mac_range(self):
        cidr = "192.168.1.0/24"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(id='1', name="public", tenant_id="make",
                       network_plugin="BASE",
                       ipam_strategy="ANY")
        network = {"network": network}
        subnet_v4 = dict(id='1', ip_version=4, next_auto_assign_ip=2,
                         cidr=cidr, first_ip=ip_network.first,
                         last_ip=ip_network.last, ip_policy=None,
                         tenant_id="fake")
        subnet_v4_info = {"subnet": subnet_v4}

        def _make_body(use_forbidden_mac_range=False):
            fix_ipv4 = dict(subnet_id=sub_v4s[0]['id'])
            port_info = \
                {"port": dict(
                    fixed_ips=[fix_ipv4], network_id=net['id'],
                    use_forbidden_mac_range=use_forbidden_mac_range
                )}
            return port_info

        with self._stubs(network, [subnet_v4_info]) as (
                net, sub_v4s, sub_v6):
            admin_ctxt = self.context.elevated()

            port = port_api.create_port(
                admin_ctxt, _make_body(use_forbidden_mac_range=True))
            port_mac = port["mac_address"]

            self.assertTrue(port_mac.startswith("AA:BB:CC"))

            with self.assertRaises(q_exc.MacAddressGenerationFailure):
                port_api.create_port(admin_ctxt,
                                     _make_body())


class QuarkFindPortsFilterByDeviceOwner(BaseFunctionalTest):
    def test_port_list_device_owner_found_returns_only_those(self):
        # create a network
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        net_mod = db_api.network_create(self.context, **network)
        # create ports
        port1 = dict(network_id=net_mod["id"], backend_key="1", device_id="1",
                     device_owner="Doge")
        port2 = dict(network_id=net_mod["id"], backend_key="1", device_id="1",
                     device_owner=port1["device_owner"])
        port3 = dict(network_id=net_mod["id"], backend_key="1", device_id="1",
                     device_owner="network:dhcp")
        port_mod1 = db_api.port_create(self.context, **port1)
        port_mod2 = db_api.port_create(self.context, **port2)
        port_mod3 = db_api.port_create(self.context, **port3)
        res = db_api.port_find(self.context, None, None, None,
                               scope=db_api.ALL,
                               device_owner=port3["device_owner"])
        self.assertTrue(len(res) == 1)
        self.assertTrue(res[0]["device_owner"] == port3["device_owner"])
        res = db_api.port_find(self.context, None, None, None,
                               scope=db_api.ALL,
                               device_owner=port1["device_owner"])
        self.assertTrue(len(res) == 2)
        self.assertTrue(res[0]["device_owner"] == res[1]["device_owner"] ==
                        port1["device_owner"])
        db_api.network_delete(self.context, net_mod)
        db_api.port_delete(self.context, port_mod1)
        db_api.port_delete(self.context, port_mod2)
        db_api.port_delete(self.context, port_mod3)


class QuarkPortFixedIPOperations(BaseFunctionalTest):

    def __init__(self, *args, **kwargs):
        super(QuarkPortFixedIPOperations, self).__init__(*args, **kwargs)
        cidr = "192.168.10.0/24"
        ip_network = netaddr.IPNetwork(cidr)
        cidr_v6 = "2001:db8::/32"
        ip_network_v6 = netaddr.IPNetwork(cidr_v6)
        # some default stuff
        network = dict(name="public", tenant_id="make",
                       network_plugin="BASE",
                       ipam_strategy="ANY")
        self.net_info = {"network": network}
        subnet_v4 = dict(ip_version=4, next_auto_assign_ip=2,
                         cidr=cidr, first_ip=ip_network.first,
                         last_ip=ip_network.last, ip_policy=None,
                         tenant_id="fake")
        subnet_v6 = dict(ip_version=6, next_auto_assign_ip=2,
                         cidr=cidr_v6, first_ip=ip_network_v6.first,
                         last_ip=ip_network_v6.last, ip_policy=None,
                         tenant_id="fake")
        self.sub_info = {"subnet": subnet_v4}
        self.sub_info_v6 = {"subnet": subnet_v6}

    @contextlib.contextmanager
    def _stubs(self, network_info, subnet_info):
        with contextlib.nested(
                mock.patch("neutron.common.rpc.get_notifier"),
                mock.patch("neutron.quota.QUOTAS.limit_check")):
            mac = {'mac_address_range': dict(cidr="AA:BB:CC")}
            self.context.is_admin = True
            macrng_api.create_mac_address_range(self.context, mac)
            self.context.is_admin = False
            network = network_api.create_network(self.context, network_info)
            subnet_info['subnet']['network_id'] = network['id']
            subnet = subnet_api.create_subnet(self.context, subnet_info)

            yield network, subnet

    def test_create_port_single_fixed_ip(self):
        with self._stubs(self.net_info, self.sub_info) as (network, subnet):
            fixed_ips = [dict(subnet_id=subnet['id'], enabled=True,
                         ip_address="192.168.10.45")]
            port = dict(port=dict(network_id=network['id'],
                                  tenant_id=self.context.tenant_id,
                                  device_id=2,
                                  fixed_ips=fixed_ips))
            expected = {'status': "ACTIVE",
                        'device_owner': None,
                        'network_id': network["id"],
                        'tenant_id': self.context.tenant_id,
                        'admin_state_up': True,
                        'fixed_ips': fixed_ips,
                        'device_id': 2}
            result = port_api.create_port(self.context, port)
            for key in expected.keys():
                self.assertEqual(result[key], expected[key],
                                 "Mismatch on %s" % key)

    def test_create_port_multiple_fixed_ipv4(self):
        with self._stubs(self.net_info, self.sub_info) as (network, subnet):
            fixed_ips = [dict(subnet_id=subnet['id'], enabled=True,
                         ip_address="192.168.10.45"),
                         dict(subnet_id=subnet['id'], enabled=True,
                         ip_address="192.168.10.199")]
            port = dict(port=dict(network_id=network['id'],
                                  tenant_id=self.context.tenant_id,
                                  device_id=2,
                                  fixed_ips=fixed_ips))
            expected = {'status': "ACTIVE",
                        'device_owner': None,
                        'network_id': network["id"],
                        'tenant_id': self.context.tenant_id,
                        'admin_state_up': True,
                        'fixed_ips': fixed_ips,
                        'device_id': 2}
            result = port_api.create_port(self.context, port)
            for key in expected.keys():
                if key != 'fixed_ips':
                    self.assertEqual(result[key], expected[key],
                                     "Mismatch on %s" % key)
            for ip in result['fixed_ips']:
                self.assertTrue(ip in expected['fixed_ips'])

    def test_create_port_multiple_fixed_ipv6(self):
        with self._stubs(self.net_info, self.sub_info_v6) as (network, subnet):
            ipv6a = "2001:db8::10"
            ipv6b = "2001:db8::15"
            fixed_ips = [dict(subnet_id=subnet['id'], enabled=True,
                         ip_address=ipv6a),
                         dict(subnet_id=subnet['id'], enabled=True,
                         ip_address=ipv6b)]
            port = dict(port=dict(network_id=network['id'],
                                  tenant_id=self.context.tenant_id,
                                  device_id=2,
                                  fixed_ips=fixed_ips))
            expected = {'status': "ACTIVE",
                        'device_owner': None,
                        'network_id': network["id"],
                        'tenant_id': self.context.tenant_id,
                        'admin_state_up': True,
                        'fixed_ips': fixed_ips,
                        'device_id': 2}
            result = port_api.create_port(self.context, port)
            for key in expected.keys():
                if key != 'fixed_ips':
                    self.assertEqual(result[key], expected[key],
                                     "Mismatch on %s" % key)
            for ip in result['fixed_ips']:
                self.assertTrue(ip in expected['fixed_ips'])

    def test_update_port_multiple_fixed_ipv4(self):
        with self._stubs(self.net_info, self.sub_info) as (network, subnet):
            fixed_ips = [dict(subnet_id=subnet['id'], enabled=True,
                         ip_address="192.168.10.45"),
                         dict(subnet_id=subnet['id'], enabled=True,
                         ip_address="192.168.10.199")]
            port = dict(port=dict(network_id=network['id'],
                                  tenant_id=self.context.tenant_id,
                                  device_id=2,
                                  fixed_ips=fixed_ips))
            expected = {'status': "ACTIVE",
                        'device_owner': None,
                        'network_id': network["id"],
                        'tenant_id': self.context.tenant_id,
                        'admin_state_up': True,
                        'fixed_ips': fixed_ips,
                        'device_id': '2'}
            result = port_api.create_port(self.context, port)

            fixed_ips = [dict(subnet_id=subnet['id'], enabled=True,
                         ip_address="192.168.10.236"),
                         dict(subnet_id=subnet['id'], enabled=True,
                         ip_address="192.168.10.42")]
            new_port = dict(port=dict(fixed_ips=fixed_ips))
            result = port_api.update_port(self.context, result['id'], new_port)
            for key in expected.keys():
                if key != 'fixed_ips':
                    self.assertEqual(result[key], expected[key],
                                     "Mismatch on %s" % key)
            for ip in result['fixed_ips']:
                self.assertTrue(ip in fixed_ips,
                                '%s not in %s' % (ip, expected['fixed_ips']))

    def test_port_show(self):
        with self._stubs(self.net_info, self.sub_info) as (network, subnet):
            fixed_ips = [dict(subnet_id=subnet['id'], enabled=True,
                         ip_address="192.168.10.45"),
                         dict(subnet_id=subnet['id'], enabled=True,
                         ip_address="192.168.10.199")]
            port = dict(port=dict(network_id=network['id'],
                                  tenant_id=self.context.tenant_id,
                                  device_id='2',
                                  fixed_ips=fixed_ips))
            expected = {'status': "ACTIVE",
                        'device_owner': None,
                        'network_id': network["id"],
                        'tenant_id': self.context.tenant_id,
                        'admin_state_up': True,
                        'fixed_ips': fixed_ips,
                        'device_id': '2'}
            result = port_api.create_port(self.context, port)
            result = port_api.get_port(self.context, result['id'])
            for key in expected.keys():
                if key != 'fixed_ips':
                    self.assertEqual(result[key], expected[key],
                                     "Mismatch on %s" % key)
            for ip in result['fixed_ips']:
                self.assertTrue(ip in fixed_ips,
                                '%s not in %s' % (ip, expected['fixed_ips']))


class QuarkAdvancedServiceCreatePort(BaseFunctionalTest):

    def __init__(self, *args, **kwargs):
        super(QuarkAdvancedServiceCreatePort, self).__init__(*args, **kwargs)
        cidr = "192.168.10.0/24"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(name="public",
                       network_plugin="BASE",
                       ipam_strategy="ANY")
        self.net_info = {"network": network}
        subnet_v4 = dict(ip_version=4, next_auto_assign_ip=2,
                         cidr=cidr, first_ip=ip_network.first,
                         last_ip=ip_network.last, ip_policy=None)
        self.sub_info = {"subnet": subnet_v4}

    @contextlib.contextmanager
    def _stubs(self, network_info, subnet_info):
        with contextlib.nested(
                mock.patch("neutron.common.rpc.get_notifier"),
                mock.patch("neutron.quota.QUOTAS.limit_check")):
            mac = {'mac_address_range': dict(cidr="AA:BB:CC")}
            self.context.is_admin = True
            macrng_api.create_mac_address_range(self.context, mac)
            self.context.is_admin = False
            # Setting context's tenant_id because this network needs to belong
            # to a regular tenant, and the network create method does not
            # care about the tenant_id on the network
            self.context.tenant_id = 'joetenant'
            network = network_api.create_network(self.context, network_info)
            subnet_info['subnet']['network_id'] = network['id']
            subnet = subnet_api.create_subnet(self.context, subnet_info)
            self.context.tenant_id = 'advsvc'
            yield network, subnet

    def test_can_create_port_with_adv_svc(self):
        with self._stubs(self.net_info, self.sub_info) as (network, subnet):
            port_info = {'port': {'network_id': network['id'],
                                  'tenant_id': 'someoneelse'}}
            self.context.is_admin = True
            self.context.is_advsvc = True
            port_mod = port_api.create_port(self.context, port_info)
            self.assertIsNotNone(port_mod['id'])
            self.assertNotEqual(port_mod['tenant_id'], network['tenant_id'])

    def test_cant_create_port_without_adv_svc(self):
        with self._stubs(self.net_info, self.sub_info) as (network, subnet):
            port_info = {'port': {'network_id': network['id'],
                                  'tenant_id': 'someoneelse'}}
            self.context.is_admin = True
            self.context.is_advsvc = False
            self.assertRaises(q_exc.NotAuthorized,
                              port_api.create_port, self.context, port_info)

    def test_cant_create_port_without_admin(self):
        with self._stubs(self.net_info, self.sub_info) as (network, subnet):
            port_info = {'port': {'network_id': network['id'],
                                  'tenant_id': 'someoneelse'}}
            self.context.is_admin = False
            self.context.is_advsvc = True
            # This is NetworkNotFound because prior to doing the authorized
            # check, quark will first attempt to retrieve the network but
            # since networks are scoped by tenant when it is not an admin,
            # it will not be found
            self.assertRaises(q_exc.NetworkNotFound,
                              port_api.create_port, self.context, port_info)



"""drop deallocated_at from mac addresses index

Revision ID: 2a116b962c95
Revises: 41837dc547ce3
Create Date: 2016-06-02 19:26:27.899610

"""

# revision identifiers, used by Alembic.
revision = '2a116b962c95'
down_revision = '41837dc547ce3'

from alembic import op


def upgrade():
    op.drop_index(op.f('ix_quark_mac_addresses_reallocation'),
                  table_name='quark_mac_addresses')
    op.create_index(op.f('ix_quark_mac_addresses_reallocation'),
                    'quark_mac_addresses', ['deallocated'],
                    unique=False)


def downgrade():
    op.drop_index(op.f('ix_quark_mac_addresses_reallocation'),
                  table_name='quark_mac_addresses')
    op.create_index(op.f('ix_quark_mac_addresses_reallocation'),
                    'quark_mac_addresses', ['deallocated', 'deallocated_at'],
                    unique=False)



# Copyright (c) 2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
import mock
from oslo_config import cfg
from quark import billing
from quark.db.models import IPAddress
from quark import network_strategy
from quark.tests import test_base
from quark import utils


class QuarkBillingBaseTest(test_base.TestBase):
    def setUp(self):
        super(QuarkBillingBaseTest, self).setUp()
        self.strategy = {"00000000-0000-0000-0000-000000000000":
                         {"bridge": "publicnet",
                          "subnets": {"4": "public_v4",
                                      "6": "public_v6"}}}
        strategy_json = json.dumps(self.strategy)
        cfg.CONF.set_override("default_net_strategy", strategy_json, "QUARK")
        network_strategy.STRATEGY.load()
        # Need to patch it here because billing has loaded before this
        # code is executed and had no config available to it.
        billing.PUBLIC_NETWORK_ID = \
            network_strategy.STRATEGY.get_public_net_id()


TENANT_ID = u'12345'
IP_ID = 'ffffffff-dddd-cccc-bbbb-aaaaaaaaaaaa'
IP_READABLE = '1.1.1.1'
SUBNET_ID = 'badc0ffe-dead-beef-c0fe-baaaaadc0ded'
PUB_NETWORK_ID = '00000000-0000-0000-0000-000000000000'


def get_fake_fixed_address():
    ipaddress = IPAddress()
    ipaddress.used_by_tenant_id = TENANT_ID
    ipaddress.version = 4
    ipaddress.id = IP_ID
    ipaddress.address_readable = IP_READABLE
    ipaddress.subnet_id = SUBNET_ID
    ipaddress.network_id = PUB_NETWORK_ID

    return ipaddress


class QuarkBillingPayloadTest(QuarkBillingBaseTest):
    """Tests for payload generation.

    This is the payload json:
    {
        'event_type': unicode(EVENT_TYPE_2_CLOUDFEEDS[event_type]),
        'tenant_id': unicode(ipaddress.used_by_tenant_id),
        'ip_address': unicode(ipaddress.address_readable),
        'subnet_id': unicode(ipaddress.subnet_id),
        'network_id': unicode(ipaddress.network_id),
        'public': True if ipaddress.network_id == PUBLIC_NETWORK_ID else False,
        'ip_version': int(ipaddress.version),
        'ip_type': unicode(ipaddress.address_type),
        'id': unicode(ipaddress.id)
    }
    """
    def setUp(self):
        super(QuarkBillingPayloadTest, self).setUp()

    def test_fixed_payload(self):
        start_time = datetime.datetime.utcnow().replace(microsecond=0) -\
            datetime.timedelta(days=1)
        end_time = datetime.datetime.utcnow().replace(microsecond=0)
        ipaddress = get_fake_fixed_address()
        ipaddress.allocated_at = start_time
        ipaddress.deallocated_at = end_time
        ipaddress.address_type = 'fixed'
        payload = billing.build_payload(ipaddress, billing.IP_EXISTS,
                                        start_time=start_time,
                                        end_time=end_time)
        self.assertEqual(payload['event_type'], billing.IP_EXISTS,
                         'event_type is wrong')
        self.assertEqual(payload['tenant_id'], TENANT_ID,
                         'tenant_id is wrong')
        self.assertEqual(payload['ip_address'], IP_READABLE,
                         'ip_address is wrong')
        self.assertEqual(payload['ip_version'], 4,
                         'ip_version should be 4')
        self.assertEqual(payload['ip_type'], 'fixed',
                         'ip_type should be fixed')
        self.assertEqual(payload['id'], IP_ID, 'ip_id is wrong')
        self.assertEqual(payload['startTime'],
                         billing.convert_timestamp(start_time),
                         'startTime is wrong')
        self.assertEqual(payload['endTime'],
                         billing.convert_timestamp(end_time),
                         'endTime is wrong')

    def test_associate_flip_payload(self):
        event_time = datetime.datetime.utcnow().replace(microsecond=0)
        ipaddress = get_fake_fixed_address()
        # allocated_at and deallocated_at could be anything for testing this
        ipaddress.allocated_at = event_time
        ipaddress.deallocated_at = event_time
        ipaddress.address_type = 'floating'
        payload = billing.build_payload(ipaddress, billing.IP_ASSOC,
                                        event_time=event_time)
        self.assertEqual(payload['event_type'], billing.IP_ASSOC,
                         'event_type is wrong')
        self.assertEqual(payload['tenant_id'], TENANT_ID, 'tenant_id is wrong')
        self.assertEqual(payload['ip_address'], IP_READABLE,
                         'ip_address is wrong')
        self.assertEqual(payload['subnet_id'], SUBNET_ID, 'subnet_id is wrong')
        self.assertEqual(payload['network_id'], PUB_NETWORK_ID,
                         'network_id is wrong')
        self.assertEqual(payload['public'], True, 'public should be true')
        self.assertEqual(payload['ip_version'], 4, 'ip_version should be 4')
        self.assertEqual(payload['ip_type'], 'floating',
                         'ip_type should be fixed')
        self.assertEqual(payload['id'], IP_ID, 'ip_id is wrong')
        self.assertEqual(payload['eventTime'],
                         billing.convert_timestamp(event_time),
                         'eventTime is wrong')

    def test_disassociate_flip_payload(self):
        event_time = datetime.datetime.utcnow().replace(microsecond=0)
        ipaddress = get_fake_fixed_address()
        # allocated_at and deallocated_at could be anything for testing this
        ipaddress.allocated_at = event_time
        ipaddress.deallocated_at = event_time
        ipaddress.address_type = 'floating'
        payload = billing.build_payload(ipaddress, billing.IP_DISASSOC,
                                        event_time=event_time)
        self.assertEqual(payload['event_type'], billing.IP_DISASSOC,
                         'event_type is wrong')
        self.assertEqual(payload['tenant_id'], TENANT_ID, 'tenant_id is wrong')
        self.assertEqual(payload['ip_address'], IP_READABLE,
                         'ip_address is wrong')
        self.assertEqual(payload['subnet_id'], SUBNET_ID, 'subnet_id is wrong')
        self.assertEqual(payload['network_id'], PUB_NETWORK_ID,
                         'network_id is wrong')
        self.assertEqual(payload['public'], True, 'public should be true')
        self.assertEqual(payload['ip_version'], 4, 'ip_version should be 4')
        self.assertEqual(payload['ip_type'], 'floating',
                         'ip_type should be fixed')
        self.assertEqual(payload['id'], IP_ID, 'ip_id is wrong')
        self.assertEqual(payload['eventTime'],
                         billing.convert_timestamp(event_time),
                         'eventTime is wrong')


class QuarkBillingEnvironmentCapabilityTest(QuarkBillingBaseTest):
    def setUp(self):
        super(QuarkBillingEnvironmentCapabilityTest, self).setUp()
        self.context = {}

    @mock.patch('neutron.common.rpc.get_notifier')
    def test_env_cap_enabled(self, notifier):
        cfg.CONF.set_override('environment_capabilities',
                              'security_groups,ip_billing',
                              'QUARK')
        ipaddress = get_fake_fixed_address()
        ipaddress.allocated_at = datetime.datetime.utcnow()
        billing.notify(self.context, billing.IP_ADD, ipaddress)
        notifier.assert_called_once_with('network')
        cfg.CONF.clear_override('environment_capabilities', 'QUARK')

    @mock.patch('neutron.common.rpc.get_notifier')
    def test_env_cap_disabled(self, notifier):
        cfg.CONF.set_override('environment_capabilities',
                              '',
                              'QUARK')
        ipaddress = get_fake_fixed_address()
        ipaddress.allocated_at = datetime.datetime.utcnow()
        billing.notify(self.context, billing.IP_ADD, ipaddress)
        self.assertFalse(notifier.called)
        cfg.CONF.clear_override('environment_capabilities', 'QUARK')

    @mock.patch('neutron.common.rpc.get_notifier')
    def test_do_not_notify_in_undo_cmd_mgr(self, notifier):
        """Wraps a call to notify in CommandManager's rollback()"""
        cfg.CONF.set_override('environment_capabilities',
                              'security_groups,ip_billing',
                              'QUARK')
        ipaddress = get_fake_fixed_address()
        ipaddress.allocated_at = datetime.datetime.utcnow()
        try:
            with utils.CommandManager().execute() as cmd_mgr:
                @cmd_mgr.do
                def f():
                    raise Exception

                @cmd_mgr.undo
                def f_undo(*args, **kwargs):
                    billing.notify(self.context, billing.IP_ADD, ipaddress,
                                   *args, **kwargs)

                f()
        except Exception:
            pass

        # notifier should NOT have been called
        self.assertFalse(notifier.called)
        cfg.CONF.clear_override('environment_capabilities', 'QUARK')



"""Populate IPPolicy.size

Revision ID: 28e55acaf366
Revises: 3d22de205729
Create Date: 2014-08-06 14:50:04.022331

"""

# revision identifiers, used by Alembic.
revision = '28e55acaf366'
down_revision = '3d22de205729'

from alembic import op
from sqlalchemy.sql import column, select, table
import netaddr
import sqlalchemy as sa

from quark.db.custom_types import INET


def upgrade():
    ip_policy = table('quark_ip_policy',
                      column('id', sa.String(length=36)),
                      column('size', INET()))
    ip_policy_cidrs = table('quark_ip_policy_cidrs',
                            column('ip_policy_id', sa.String(length=36)),
                            column('cidr', sa.String(length=64)))
    connection = op.get_bind()

    # 1. Retrieve all ip_policy_cidr rows.
    results = connection.execute(
        select([ip_policy_cidrs.c.ip_policy_id, ip_policy_cidrs.c.cidr])
    ).fetchall()

    # 2. Determine IPSet for each IP Policy.
    ipp = dict()
    for ip_policy_id, cidr in results:
        if ip_policy_id not in ipp:
            ipp[ip_policy_id] = netaddr.IPSet()
        ipp[ip_policy_id].add(cidr)

    # 3. Populate size for each IP Policy.
    for ip_policy_id in ipp:
        connection.execute(ip_policy.update().values(
            size=ipp[ip_policy_id].size).where(
                ip_policy.c.id == ip_policy_id))


def downgrade():
    raise NotImplementedError()



import datetime

import mock
import netaddr
from neutron.db import api as neutron_db_api
from oslo_config import cfg

from quark.db import models
from quark import ip_availability as ip_avail
from quark.tests.functional.base import BaseFunctionalTest


EPOCH = datetime.date(1970, 1, 1)


class QuarkIpAvailabilityBaseFunctionalTest(BaseFunctionalTest):
    def setUp(self):
        super(QuarkIpAvailabilityBaseFunctionalTest, self).setUp()
        self.connection = neutron_db_api.get_engine().connect()
        self.networks = models.BASEV2.metadata.tables["quark_networks"]
        self.subnets = models.BASEV2.metadata.tables["quark_subnets"]
        self.ip_policy = models.BASEV2.metadata.tables[
            "quark_ip_policy"]
        self.ip_policy_cidr = models.BASEV2.metadata.tables[
            "quark_ip_policy_cidrs"]
        self.ip_addresses = models.BASEV2.metadata.tables[
            "quark_ip_addresses"]
        self.locks = models.BASEV2.metadata.tables[
            "quark_locks"]
        self.default_kwargs = {
            "network_id": "00000000-0000-0000-0000-000000000000",
            "ip_version": 4}

    def _insert_ip_policy(self, id=0, excludes=None):
        if not excludes:
            excludes = (0, 255)
        self.connection.execute(
            self.ip_policy.insert(),
            id=id, size=len(excludes), created_at=EPOCH)
        self.connection.execute(
            self.ip_policy_cidr.insert(),
            [dict(ip_policy_id=id, first_ip=x, last_ip=x, created_at=EPOCH)
             for x in excludes])

    def _insert_network(self, id="00000000-0000-0000-0000-000000000000"):
        self.connection.execute(self.networks.insert(),
                                id=id, created_at=EPOCH)

    def _insert_subnet(self,
                       do_not_use=0,
                       id=0,
                       network_id="00000000-0000-0000-0000-000000000000",
                       cidr="0.0.0.0/24",
                       segment_id="region-cell",
                       ip_policy_id=0,
                       ip_version=4):
        self.connection.execute(
            self.subnets.insert(),
            do_not_use=do_not_use,
            _cidr=cidr,
            network_id=network_id,
            ip_version=ip_version,
            segment_id=segment_id,
            id=id,
            ip_policy_id=ip_policy_id,
            created_at=EPOCH)

    def _insert_ip_address(self,
                           address=1,
                           address_readable="0.0.0.1",
                           subnet_id=0,
                           deallocated=0,
                           deallocated_at=None,
                           lock_id=None):
        self.connection.execute(
            self.ip_addresses.insert(),
            address=address,
            address_readable=address_readable,
            subnet_id=subnet_id,
            _deallocated=deallocated,
            deallocated_at=deallocated_at,
            lock_id=lock_id,
            created_at=EPOCH)

    def _insert_lock(self, id=0, type="ip_address"):
        self.connection.execute(
            self.locks.insert(),
            created_at=EPOCH,
            id=id,
            type=type)

    def _default(self):
        self._insert_ip_policy()
        self._insert_network()
        self._insert_subnet()
        self._insert_ip_address()

    def _do_not_use_None(self):
        self._insert_ip_policy()
        self._insert_network()
        self._insert_subnet(do_not_use=None)
        self._insert_ip_address()

    def _do_not_use_1(self):
        self._insert_ip_policy()
        self._insert_network()
        self._insert_subnet(do_not_use=1)
        self._insert_ip_address()

    def _no_ip_addresses(self):
        self._insert_ip_policy()
        self._insert_network()
        self._insert_subnet()

    def _no_ip_addresses_no_ip_policy(self):
        self._insert_network()
        self._insert_subnet(ip_policy_id=None)

    def _no_ip_policy(self, utcnow_patch):
        self._insert_network()

        base = datetime.datetime(2015, 2, 13)
        utcnow_patch.return_value = base
        delta = datetime.timedelta(seconds=cfg.CONF.QUARK.ipam_reuse_after)
        reuse_window = base - delta
        epsilon = datetime.timedelta(seconds=1)

        address_readable = {0: "0.0.0.0",
                            1: "0.0.0.1",
                            255: "0.0.0.255"}
        subnet_id = 0
        for lock in (None, 1):
            for deallocated in (None, 0, 1):
                for deallocated_at in (None,
                                       reuse_window - epsilon,
                                       reuse_window,
                                       reuse_window + epsilon):
                    for address in address_readable:
                        self._insert_subnet(id=subnet_id, ip_policy_id=None)
                        lock_id = None
                        if lock:
                            lock_id = subnet_id
                            self._insert_lock(id=lock_id)
                        self._insert_ip_address(
                            subnet_id=subnet_id,
                            address=address,
                            deallocated=deallocated,
                            deallocated_at=deallocated_at,
                            address_readable=address_readable[address],
                            lock_id=lock_id)
                        subnet_id += 1

    def _with_ip_policy(self, utcnow_patch):
        self._insert_network()
        self._insert_ip_policy()

        base = datetime.datetime(2015, 2, 13)
        utcnow_patch.return_value = base
        delta = datetime.timedelta(seconds=cfg.CONF.QUARK.ipam_reuse_after)
        reuse_window = base - delta
        epsilon = datetime.timedelta(seconds=1)

        address_readable = {0: "0.0.0.0",
                            1: "0.0.0.1",
                            255: "0.0.0.255"}
        subnet_id = 0
        for lock in (None, 1):
            for deallocated in (None, 0, 1):
                for deallocated_at in (None,
                                       reuse_window - epsilon,
                                       reuse_window,
                                       reuse_window + epsilon):
                    for address in address_readable:
                        self._insert_subnet(id=subnet_id)
                        lock_id = None
                        if lock:
                            lock_id = subnet_id
                            self._insert_lock(id=lock_id)
                        self._insert_ip_address(
                            subnet_id=subnet_id,
                            address=address,
                            deallocated=deallocated,
                            deallocated_at=deallocated_at,
                            address_readable=address_readable[address])
                        subnet_id += 1


class QuarkIpAvailabilityFunctionalTest(QuarkIpAvailabilityBaseFunctionalTest):
    def test_empty(self):
        output = ip_avail.get_ip_availability(**self.default_kwargs)
        self.assertEqual(output["used"], {})
        self.assertEqual(output["unused"], {})

    def test_default(self):
        self._default()
        output = ip_avail.get_ip_availability(**self.default_kwargs)
        self.assertEqual(output["used"], {"region-cell": 1})
        self.assertEqual(output["unused"], {"region-cell": 253})

    def test_do_not_use_None(self):
        self._do_not_use_None()
        output = ip_avail.get_ip_availability(**self.default_kwargs)
        self.assertEqual(output["used"], {"region-cell": 1})
        self.assertEqual(output["unused"], {"region-cell": 253})

    def test_do_not_use_1(self):
        self._do_not_use_1()
        output = ip_avail.get_ip_availability(**self.default_kwargs)
        self.assertEqual(output["used"], dict())
        self.assertEqual(output["unused"], dict())

    def test_no_ip_addresses(self):
        self._no_ip_addresses()
        output = ip_avail.get_ip_availability(**self.default_kwargs)
        self.assertEqual(output["used"], {"region-cell": 0})
        self.assertEqual(output["unused"], {"region-cell": 254})

    def test_no_ip_addresses_no_ip_policy(self):
        self._no_ip_addresses_no_ip_policy()
        output = ip_avail.get_ip_availability(**self.default_kwargs)
        self.assertEqual(output["used"], {"region-cell": 0})
        self.assertEqual(output["unused"], {"region-cell": 256})

    @mock.patch("quark.ip_availability.timeutils.utcnow")
    def test_no_ip_policy(self, utcnow_patch):
        self._no_ip_policy(utcnow_patch)
        output = ip_avail.get_ip_availability(**self.default_kwargs)
        self.assertEqual(output["used"], {"region-cell": 63})
        self.assertEqual(output["unused"], {"region-cell": 256 * 72 - 63})

    @mock.patch("quark.ip_availability.timeutils.utcnow")
    def test_with_ip_policy(self, utcnow_patch):
        self._with_ip_policy(utcnow_patch)
        output = ip_avail.get_ip_availability(**self.default_kwargs)
        self.assertEqual(output["used"], {"region-cell": 50})
        self.assertEqual(output["unused"], {"region-cell": 254 * 72 - 50})


class QuarkIpAvailabilityFilterTest(QuarkIpAvailabilityBaseFunctionalTest):
    def setUp(self):
        super(QuarkIpAvailabilityFilterTest, self).setUp()

        subnet_id = 0
        for network_id in (0, 1):
            self._insert_network(id=network_id)
            for segment_id in (0, 1):
                for ip_version in (4, 6):
                    ip_policy_id = subnet_id
                    if ip_version == 6:
                        net6 = netaddr.IPNetwork('::ffff:0.0.0.0/120')
                        excludes = (net6.first, net6.last)
                        cidr = str(net6.cidr)
                        address = net6.first + 1
                    else:
                        cidr = "0.0.0.0/24"
                        address = 1
                        excludes = (0, 255)

                    self._insert_ip_policy(
                        id=ip_policy_id,
                        excludes=excludes)
                    self._insert_subnet(
                        id=subnet_id,
                        network_id=network_id,
                        segment_id=segment_id,
                        ip_policy_id=ip_policy_id,
                        cidr=cidr,
                        ip_version=ip_version)
                    self._insert_ip_address(
                        subnet_id=subnet_id,
                        address=address,
                        address_readable=str(netaddr.IPAddress(address)))
                    subnet_id += 1

    def test_all_None(self):
        kwargs = {}
        output = ip_avail.get_ip_availability(**kwargs)
        self.assertEqual(output["used"], {"0": 4, "1": 4})
        self.assertEqual(output["unused"], {"0": 253 * 4, "1": 253 * 4})

    def test_network_id_specific(self):
        kwargs = {"network_id": 0}
        output = ip_avail.get_ip_availability(**kwargs)
        self.assertEqual(output["used"], {"0": 2, "1": 2})
        self.assertEqual(output["unused"], {"0": 253 * 2, "1": 253 * 2})

    def test_network_id_many(self):
        kwargs = {"network_id": [0, 1]}
        output = ip_avail.get_ip_availability(**kwargs)
        self.assertEqual(output["used"], {"0": 4, "1": 4})
        self.assertEqual(output["unused"], {"0": 253 * 4, "1": 253 * 4})

    def test_segment_id_specific(self):
        kwargs = {"segment_id": 0}
        output = ip_avail.get_ip_availability(**kwargs)
        self.assertEqual(output["used"], {"0": 4})
        self.assertEqual(output["unused"], {"0": 253 * 4})

    def test_segment_id_many(self):
        kwargs = {"segment_id": [0, 1]}
        output = ip_avail.get_ip_availability(**kwargs)
        self.assertEqual(output["used"], {"0": 4, "1": 4})
        self.assertEqual(output["unused"], {"0": 253 * 4, "1": 253 * 4})

    def test_ip_version_4(self):
        kwargs = {"ip_version": 4}
        output = ip_avail.get_ip_availability(**kwargs)
        self.assertEqual(output["used"], {"0": 2, "1": 2})
        self.assertEqual(output["unused"], {"0": 253 * 2, "1": 253 * 2})

    def test_ip_version_6(self):
        kwargs = {"ip_version": 6}
        output = ip_avail.get_ip_availability(**kwargs)
        self.assertEqual(output["used"], {"0": 2, "1": 2})
        self.assertEqual(output["unused"], {"0": 253 * 2, "1": 253 * 2})

    def test_ip_version_many(self):
        kwargs = {"ip_version": [4, 6]}
        output = ip_avail.get_ip_availability(**kwargs)
        self.assertEqual(output["used"], {"0": 4, "1": 4})
        self.assertEqual(output["unused"], {"0": 253 * 4, "1": 253 * 4})

    def test_subnet_id_specific(self):
        kwargs = {"subnet_id": 3}
        output = ip_avail.get_ip_availability(**kwargs)
        self.assertEqual(output["used"], {"1": 1})
        self.assertEqual(output["unused"], {"1": 253})

    def test_subnet_id_many(self):
        kwargs = {"subnet_id": [3, 4, 5]}
        output = ip_avail.get_ip_availability(**kwargs)
        self.assertEqual(output["used"], {"0": 2, "1": 1})
        self.assertEqual(output["unused"], {"0": 253 * 2, "1": 253})



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.

import contextlib

import mock
from mock import patch
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
import webob

from quark.db import ip_types
from quark.db import models
from quark import exceptions as q_exc
from quark.plugin_modules import ip_addresses
from quark.tests import test_quark_plugin


def _port_associate_stub(context, ports, address, **kwargs):
    for port in ports:
        assoc = models.PortIpAssociation()
        assoc.port_id = port.id
        assoc.ip_address_id = address.id
        assoc.port = port
        # NOTE(thomasem): This causes address['associations'] to gain this
        # PortIpAssocation instance.
        assoc.ip_address = address
        assoc.enabled = address.address_type == "fixed"
    return address


def _port_disassociate_stub(context, ports, address):
    port_ids = [port.id for port in ports]
    for idx, assoc in enumerate(address['associations']):
        if assoc.port_id in port_ids:
            address.associations.pop(idx)
    return address


def _ip_deallocate_stub(context, address):
    address['deallocated'] = 1
    address['address_type'] = None


class TestIpAddresses(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, port, addr):
        port_model = None
        addr_model = None
        if port:
            port_model = models.Port()
            port_model.update(port)
        if addr:
            addr_model = models.IPAddress()
            addr_model.update(addr)

        def _alloc_ip(context, new_addr, net_id, port_m, *args, **kwargs):
            new_addr.extend([addr_model])

        with contextlib.nested(
            mock.patch("quark.db.api.network_find"),
            mock.patch("quark.db.api.port_find"),
            mock.patch(
                "quark.plugin_modules.ip_addresses.ipam_driver"),
            mock.patch(
                "quark.plugin_modules.ip_addresses.db_api"
                ".port_associate_ip"),
            mock.patch(
                "quark.plugin_modules.ip_addresses"
                ".validate_and_fetch_segment")
        ) as (net_f, port_find, mock_ipam, mock_port_associate_ip, validate):
            port_find.return_value = port_model
            mock_ipam.allocate_ip_address.side_effect = _alloc_ip
            mock_port_associate_ip.side_effect = _port_associate_stub
            yield

    def test_create_ip_address_by_network_and_device(self):
        old_cfg = cfg.CONF.QUARK.ipaddr_allow_fixed_ip
        cfg.CONF.set_override('ipaddr_allow_fixed_ip', True, "QUARK")
        port = dict(id=1, network_id=2, ip_addresses=[])
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4, used_by_tenant_id=1)
        with self._stubs(port=port, addr=ip):
            ip_address = {"network_id": ip["network_id"],
                          "version": 4, 'device_ids': [2]}
            response = self.plugin.create_ip_address(
                self.context, dict(ip_address=ip_address))
            self.assertIsNotNone(response["id"])
            self.assertEqual(response["network_id"], ip_address["network_id"])
            self.assertEqual(response["port_ids"], [port["id"]])
            self.assertEqual(response["subnet_id"], ip["subnet_id"])
            self.assertEqual(response['type'], None)
            self.assertEqual(response["version"], 4)
            self.assertEqual(response["ip_address"], "192.168.1.100")
            self.assertEqual(response["tenant_id"], 1)
        cfg.CONF.set_override('ipaddr_allow_fixed_ip', old_cfg, "QUARK")

    def test_create_ip_address_with_port(self):
        old_cfg = cfg.CONF.QUARK.ipaddr_allow_fixed_ip
        cfg.CONF.set_override('ipaddr_allow_fixed_ip', True, "QUARK")
        port = dict(id=1, network_id=2, ip_addresses=[])
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4)
        with self._stubs(port=port, addr=ip):
            ip_address = dict(port_ids=[port["id"]])
            ip_address['version'] = 4
            ip_address['network_id'] = 2
            response = self.plugin.create_ip_address(
                self.context, dict(ip_address=ip_address))

            self.assertIsNotNone(response['id'])
            self.assertEqual(response['network_id'], ip["network_id"])
            self.assertEqual(response['port_ids'], [port["id"]])
            self.assertEqual(response['subnet_id'], ip['id'])
        cfg.CONF.set_override('ipaddr_allow_fixed_ip', old_cfg, "QUARK")

    def test_fail_create_ip_address_with_port_when_disallowed(self):
        old_cfg = cfg.CONF.QUARK.ipaddr_allow_fixed_ip
        cfg.CONF.set_override('ipaddr_allow_fixed_ip', False, "QUARK")
        port = dict(id=1, network_id=2, ip_addresses=[])
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4)
        with self._stubs(port=port, addr=ip):
            ip_address = dict(port_ids=[port["id"]])
            ip_address['version'] = 4
            ip_address['network_id'] = 2
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.create_ip_address(self.context,
                                              dict(ip_address=ip_address))
        cfg.CONF.set_override('ipaddr_allow_fixed_ip', old_cfg, "QUARK")

    def test_create_ip_address_by_device_no_network_fails(self):
        with self._stubs(port={}, addr=None):
            ip_address = dict(device_ids=[4], version=4)
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.create_ip_address(self.context,
                                              dict(ip_address=ip_address))

    def test_create_ip_address_invalid_missing_port_and_device_list(self):
        with self._stubs(port=None, addr=None):
            with self.assertRaises(n_exc.BadRequest):
                ip_address = {'ip_address': {'network_id': 'fake',
                                             'version': 4}}
                self.plugin.create_ip_address(self.context, ip_address)

    def test_create_ip_address_invalid_port(self):
        old_cfg = cfg.CONF.QUARK.ipaddr_allow_fixed_ip
        cfg.CONF.set_override('ipaddr_allow_fixed_ip', True, "QUARK")
        with self._stubs(port=None, addr=None):
            with self.assertRaises(n_exc.PortNotFoundOnNetwork):
                ip_address = {
                    'ip_address': {
                        'port_ids': ['fake'],
                        'version': 4,
                        'network_id': 'fake'
                    }
                }
                self.plugin.create_ip_address(self.context, ip_address)
        cfg.CONF.set_override('ipaddr_allow_fixed_ip', old_cfg, "QUARK")


@mock.patch("quark.plugin_modules.ip_addresses.v")
@mock.patch("quark.plugin_modules.ip_addresses.ipam_driver")
@mock.patch("quark.plugin_modules.ip_addresses.db_api")
class TestQuarkSharedIPAddressCreate(test_quark_plugin.TestQuarkPlugin):
    def setUp(self):
        super(TestQuarkSharedIPAddressCreate, self).setUp()
        patcher = patch('quark.plugin_modules.ip_addresses.'
                        'validate_and_fetch_segment')
        segment_check = patcher.start()
        segment_check.return_value = None
        self.addCleanup(patcher.stop)

    def _alloc_stub(self, ip_model):
        def _alloc_ip(context, addr, *args, **kwargs):
            addr.append(ip_model)
        return _alloc_ip

    def test_create_ip_address_calls_port_associate_ip(self, mock_dbapi,
                                                       mock_ipam, *args):
        old_cfg = cfg.CONF.QUARK.ipaddr_allow_fixed_ip
        cfg.CONF.set_override('ipaddr_allow_fixed_ip', True, "QUARK")
        port = dict(id=1, network_id=2, ip_addresses=[])
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4, tenant_id=1)
        port_model = models.Port()
        port_model.update(port)
        ip_model = models.IPAddress()
        ip_model.update(ip)

        mock_dbapi.port_find.return_value = port_model
        mock_ipam.allocate_ip_address.side_effect = (
            self._alloc_stub(ip_model))
        ip_address = {"network_id": ip["network_id"],
                      "version": 4, 'device_ids': [2],
                      "port_ids": [1]}

        self.plugin.create_ip_address(self.context,
                                      dict(ip_address=ip_address))
        mock_dbapi.port_associate_ip.assert_called_once_with(
            self.context, [port_model], ip_model)
        cfg.CONF.set_override('ipaddr_allow_fixed_ip', old_cfg, "QUARK")

    def test_create_ip_address_address_type_shared(self, mock_dbapi, mock_ipam,
                                                   *args):
        cfg.CONF.set_override('ipam_reuse_after', 100, "QUARK")
        ports = [dict(id=1, network_id=2, ip_addresses=[]),
                 dict(id=2, network_id=2, ip_addresses=[])]
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4, tenant_id=1)
        port_models = [models.Port(**p) for p in ports]
        ip_model = models.IPAddress()
        ip_model.update(ip)
        mock_dbapi.port_find.side_effect = port_models
        mock_ipam.allocate_ip_address.side_effect = (
            self._alloc_stub(ip_model))

        ip_address = {"network_id": ip["network_id"],
                      "version": 4, 'device_ids': [2],
                      "port_ids": [pm.id for pm in port_models]}
        self.plugin.create_ip_address(self.context,
                                      dict(ip_address=ip_address))
        # NOTE(thomasem): Having to assert that [ip_model] was passed instead
        # of an empty list due to the expected behavior of this method being
        # that it mutates the passed in list. So, after it's run, the list
        # has already been mutated and it's a reference to that list that
        # we're checking. This method ought to be changed to return the new
        # IP and let the caller mutate the list, not the other way around.
        mock_ipam.allocate_ip_address.assert_called_once_with(
            self.context, [ip_model], ip['network_id'], None, 100,
            version=ip_address['version'], ip_addresses=[],
            address_type="shared", segment_id=None)

    def test_create_ip_address_address_type_fixed(self, mock_dbapi, mock_ipam,
                                                  *args):
        cfg.CONF.set_override('ipam_reuse_after', 100, "QUARK")
        old_cfg = cfg.CONF.QUARK.ipaddr_allow_fixed_ip
        cfg.CONF.set_override('ipaddr_allow_fixed_ip', True, "QUARK")
        ports = [dict(id=1, network_id=2, ip_addresses=[])]
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4, tenant_id=1)
        port_models = [models.Port(**p) for p in ports]
        ip_model = models.IPAddress()
        ip_model.update(ip)
        mock_dbapi.port_find.side_effect = port_models
        mock_ipam.allocate_ip_address.side_effect = (
            self._alloc_stub(ip_model))

        ip_address = {"network_id": ip["network_id"],
                      "version": 4, 'device_ids': [2],
                      "port_ids": [pm.id for pm in port_models]}
        self.plugin.create_ip_address(self.context,
                                      dict(ip_address=ip_address))
        # NOTE(thomasem): Having to assert that [ip_model] was passed instead
        # of an empty list due to the expected behavior of this method being
        # that it mutates the passed in list. So, after it's run, the list
        # has already been mutated and it's a reference to that list that
        # we're checking. This method ought to be changed to return the new
        # IP and let the caller mutate the list, not the other way around.
        mock_ipam.allocate_ip_address.assert_called_once_with(
            self.context, [ip_model], ip['network_id'], None, 100,
            version=ip_address['version'], ip_addresses=[],
            segment_id=None, address_type="fixed")
        cfg.CONF.set_override('ipaddr_allow_fixed_ip', old_cfg, "QUARK")


class TestQuarkSharedIPAddressPortsValid(test_quark_plugin.TestQuarkPlugin):
    def test_validate_ports_on_network_raise_segment(self):
        mock_ports = [models.Port(id="1", network_id="2"),
                      models.Port(id="2", network_id="2")]
        mock_subnets = [models.Subnet(id="1", segment_id="2"),
                        models.Subnet(id="2", segment_id="3")]
        for i, subnet in enumerate(mock_subnets):
            mock_address = models.IPAddress(id="2", network_id="2")
            mock_address.subnet = subnet
            mock_ports[i].ip_addresses.append(mock_address)

        with self.assertRaises(n_exc.BadRequest):
            ip_addresses.validate_and_fetch_segment(
                mock_ports, "2")

    def test_validate_ports_on_network_raise_segment_multiple_ips(self):
        mock_ports = [models.Port(id="1", network_id="2"),
                      models.Port(id="2", network_id="2")]
        mock_subnets = [models.Subnet(id="1", segment_id="2"),
                        models.Subnet(id="2", segment_id="3")]
        for i, subnet in enumerate(mock_subnets):
            mock_address = models.IPAddress(id="2", network_id="2")
            mock_address.subnet = subnet
            for x in xrange(i + 1):
                mock_ports[x].ip_addresses.append(mock_address)

        with self.assertRaises(n_exc.BadRequest):
            ip_addresses.validate_and_fetch_segment(
                mock_ports, "2")

    def test_validate_ports_on_network_raise_network(self):
        mock_ports = [models.Port(id="1", network_id="2"),
                      models.Port(id="2", network_id="3")]
        mock_addresses = [models.IPAddress(id="1", network_id="2"),
                          models.IPAddress(id="2", network_id="3")]

        for i, ip_address in enumerate(mock_addresses):
            ip_address.subnet = models.Subnet(id="1", segment_id="2")
            mock_ports[i].ip_addresses.append(ip_address)

        with self.assertRaises(n_exc.BadRequest):
            ip_addresses.validate_and_fetch_segment(
                mock_ports, "2")

    def test_validate_ports_on_network_valid(self):
        mock_ports = [models.Port(id="1", network_id="2"),
                      models.Port(id="2", network_id="2")]
        for p in mock_ports:
            p.ip_addresses.append(models.IPAddress(id="1", network_id="2"))
            p.ip_addresses[-1].subnet = models.Subnet(id="1", segment_id="1")

        r = ip_addresses.validate_and_fetch_segment(
            mock_ports, "2")
        self.assertEqual(r, "1")


class TestQuarkSharedIPAddress(test_quark_plugin.TestQuarkPlugin):
    def test_shared_ip_request(self):
        ip_address_mock = {"ip_address": {"port_ids": [1, 2, 3]}}
        r = ip_addresses._shared_ip_request(ip_address_mock)
        self.assertTrue(r)

    def test_shared_ip_request_false(self):
        ip_address_mock = {"ip_address": {"port_ids": [1]}}
        r = ip_addresses._shared_ip_request(ip_address_mock)
        self.assertFalse(r)

    def test_can_be_shared(self):
        mock_address = models.IPAddress(id="1", address=3232235876,
                                        address_readable="192.168.1.100",
                                        subnet_id="1", network_id="2",
                                        version=4, used_by_tenant_id="1")
        mock_assocs = []
        for x in xrange(3):
            assoc = models.PortIpAssociation()
            assoc.ip_address_id = mock_address.id
            assoc.ip_address = mock_address
            assoc.enabled = [False, False, False][x]
            mock_assocs.append(assoc)
        r = ip_addresses._can_be_shared(mock_address)
        self.assertTrue(r)

    def test_can_be_shared_false(self):
        mock_address = models.IPAddress(id="1", address=3232235876,
                                        address_readable="192.168.1.100",
                                        subnet_id="1", network_id="2",
                                        version=4, used_by_tenant_id="1")
        mock_assocs = []
        for x in xrange(3):
            assoc = models.PortIpAssociation()
            assoc.ip_address_id = mock_address.id
            assoc.ip_address = mock_address
            assoc.enabled = [False, True, False][x]
            mock_assocs.append(assoc)
        r = ip_addresses._can_be_shared(mock_address)
        self.assertFalse(r)

    @mock.patch("quark.plugin_modules.ip_addresses._shared_ip_request")
    @mock.patch("quark.plugin_modules.ip_addresses._can_be_shared")
    def test_raise_if_shared_and_enabled(self, can_be_shared_mock,
                                         shared_ip_request_mock):
        can_be_shared_mock.return_value = False
        shared_ip_request_mock.return_value = True
        obj = mock.MagicMock()
        with self.assertRaises(n_exc.BadRequest):
            ip_addresses._raise_if_shared_and_enabled(obj, obj)

    @mock.patch("quark.plugin_modules.ip_addresses._shared_ip_request")
    @mock.patch("quark.plugin_modules.ip_addresses._can_be_shared")
    def test_raise_if_shared_and_enabled_noraise(self, can_be_shared_mock,
                                                 shared_ip_request_mock):
        can_be_shared_mock.return_value = True
        shared_ip_request_mock.return_value = True
        obj = mock.MagicMock()
        r = ip_addresses._raise_if_shared_and_enabled(obj, obj)
        self.assertEqual(r, None)

    @mock.patch("quark.plugin_modules.ip_addresses._shared_ip_request")
    @mock.patch("quark.plugin_modules.ip_addresses._can_be_shared")
    def test_raise_if_shared_and_enabled_fixed_request(self,
                                                       can_be_shared_mock,
                                                       shared_ip_request_mock):
        can_be_shared_mock.return_value = True
        shared_ip_request_mock.return_value = False
        obj = mock.MagicMock()
        r = ip_addresses._raise_if_shared_and_enabled(obj, obj)
        self.assertEqual(r, None)

    @mock.patch("quark.plugin_modules.ip_addresses._shared_ip_request")
    @mock.patch("quark.plugin_modules.ip_addresses._can_be_shared")
    def test_raise_if_shared_and_enabled_fixed_request_and_not_shareable(
            self, can_be_shared_mock, shared_ip_request_mock):
        can_be_shared_mock.return_value = False
        shared_ip_request_mock.return_value = False
        obj = mock.MagicMock()
        r = ip_addresses._raise_if_shared_and_enabled(obj, obj)
        self.assertEqual(r, None)


class TestQuarkUpdateIPAddress(test_quark_plugin.TestQuarkPlugin):

    @contextlib.contextmanager
    def _stubs(self, ports, addr, addr_ports=False):
        port_models = []
        addr_model = None

        for port in ports:
            port_model = models.Port()
            port_model.update(port)
            port_models.append(port_model)
        if addr:
            addr_model = models.IPAddress()
            addr_model.update(addr)
            if addr_ports:
                addr_model.ports = port_models

        db_mod = "quark.db.api"
        with contextlib.nested(
            mock.patch("%s.port_find" % db_mod),
            mock.patch("%s.ip_address_find" % db_mod),
            mock.patch("%s.port_associate_ip" % db_mod),
            mock.patch("%s.port_disassociate_ip" % db_mod),
            mock.patch("quark.plugin_modules.ip_addresses"
                       ".validate_and_fetch_segment"),
            mock.patch("quark.plugin_modules.ip_addresses.ipam_driver")
        ) as (port_find, ip_find, port_associate_ip,
              port_disassociate_ip, val, mock_ipam):
            port_find.return_value = port_models
            ip_find.return_value = addr_model
            port_associate_ip.side_effect = _port_associate_stub
            port_disassociate_ip.side_effect = _port_disassociate_stub
            mock_ipam.deallocate_ip_address.side_effect = (
                _ip_deallocate_stub)
            yield

    def test_update_ip_address_does_not_exist(self):
        with self._stubs(ports=[], addr=None):
            with self.assertRaises(n_exc.NotFound):
                self.plugin.update_ip_address(self.context,
                                              'no_ip_address_id',
                                              {'ip_address': {'port_ids': []}})

    def test_update_ip_address_port_not_found(self):
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4)
        with self._stubs(ports=[], addr=ip):
            with self.assertRaises(n_exc.NotFound):
                ip_address = {'ip_address': {'port_ids': ['fake']}}
                self.plugin.update_ip_address(self.context,
                                              ip["id"],
                                              ip_address)

    def test_update_ip_address_specify_ports(self):
        port = dict(id=1, network_id=2, ip_addresses=[])
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4)
        with self._stubs(ports=[port], addr=ip):
            ip_address = {'ip_address': {'port_ids': [port['id']],
                                         'network_id': 2}}
            response = self.plugin.update_ip_address(self.context,
                                                     ip['id'],
                                                     ip_address)
            self.assertEqual(response['port_ids'], [port['id']])

    def test_bad_request_fixed_update_multiple_ports(self):
        port1 = dict(id=1, network_id=2, ip_addresses=[])
        port2 = dict(id=2, network_id=2, ip_addresses=[])
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4,
                  address_type=ip_types.FIXED)
        with self._stubs(ports=[port1, port2], addr=ip):
            ip_address = {'ip_address': {'port_ids': [port1['id'],
                                                      port2['id']],
                                         'network_id': 2}}
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.update_ip_address(self.context, ip['id'],
                                              ip_address)

    def _create_patch(self, path):
        patcher = patch(path)
        mocked = patcher.start()
        self.addCleanup(patcher.stop)
        return mocked

    def test_update_ip_address_update_deallocated_at(self):
        port = dict(id=1, network_id=2, ip_addresses=[])
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4, deallocated=1,
                  deallocated_at='2020-01-01 00:00:00')

        path = 'quark.plugin_modules.ip_addresses'
        lookup = self._create_patch('%s._get_deallocated_override' % path)

        with self._stubs(ports=[port], addr=ip):
            ip_address = {'ip_address': {'reset_allocation_time': True}}
            self.plugin.update_ip_address(self.admin_context, ip['id'],
                                          ip_address)
            self.assertTrue(lookup.called)

    def test_update_ip_address_update_deallocated_at_not_deallocated(self):
        port = dict(id=1, network_id=2, ip_addresses=[])
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4, deallocated=0,
                  deallocated_at='2020-01-01 00:00:00')

        path = 'quark.plugin_modules.ip_addresses'
        lookup = self._create_patch('%s._get_deallocated_override' % path)

        with self._stubs(ports=[port], addr=ip):
            ip_address = {'ip_address': {'reset_allocation_time': True}}
            self.plugin.update_ip_address(self.admin_context, ip['id'],
                                          ip_address)
            self.assertFalse(lookup.called)

    def test_update_ip_address_update_deallocated_at_not_admin(self):
        port = dict(id=1, network_id=2, ip_addresses=[])
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4, deallocated=1,
                  deallocated_at='2020-01-01 00:00:00')

        with self._stubs(ports=[port], addr=ip):
            ip_address = {'ip_address': {'reset_allocation_time': True}}
            with self.assertRaises(webob.exc.HTTPForbidden):
                self.plugin.update_ip_address(self.context, ip['id'],
                                              ip_address)

    def test_update_ip_address_no_ports(self):
        port = dict(id=1, network_id=2, ip_addresses=[])
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4)
        with self._stubs(ports=[port], addr=ip):
            ip_address = {'ip_address': {}}
            response = self.plugin.update_ip_address(self.context,
                                                     ip['id'],
                                                     ip_address)
            self.assertEqual(response['port_ids'], [])

    def test_update_ip_address_empty_ports_does_not_delete_but_errors(self):
        port = dict(id=1, network_id=2, ip_addresses=[])
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4)
        with self._stubs(ports=[port], addr=ip, addr_ports=True):
            ip_address = {'ip_address': {'port_ids': []}}
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.update_ip_address(self.context, ip['id'],
                                              ip_address)


class TestQuarkGetIpAddress(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, ips, ports):
        with mock.patch("quark.db.api.ip_address_find") as ip_find:
            port_models = []
            for port in ports:
                p = models.Port()
                p.update(port)
                port_models.append(p)
                if ips:
                    version = ips.pop("version")
                    ip_mod = models.IPAddress()
                    ip_mod.update(ips)
                    ip_mod.version = version
                    ip_mod.ports = port_models
                    # Set up Port to IP associations
                    assoc = models.PortIpAssociation()
                    assoc.port = p
                    assoc.port_id = p.id
                    assoc.ip_address = ip_mod
                    assoc.ip_address_id = ip_mod.id
                    ip_mod.associations.append(assoc)
                    ip_find.return_value = ip_mod
                else:
                    ip_find.return_value = ips
            yield

    def test_get_ip_address(self):
        port = dict(id=100)
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4)
        with self._stubs(ips=ip, ports=[port]):
            res = self.plugin.get_ip_address(self.context, 1)
            self.assertEqual(ip["id"], res["id"])
            self.assertEqual(ip["subnet_id"], res["subnet_id"])
            self.assertEqual(ip["address_readable"], res["ip_address"])
            self.assertEqual(res["port_ids"][0], port["id"])

    def test_get_ip_address_no_ip_fails(self):
        port = dict(id=100)
        with self._stubs(ips=None, ports=[port]):
            with self.assertRaises(q_exc.IpAddressNotFound):
                self.plugin.get_ip_address(self.context, 1)


class TestQuarkGetIpAddresses(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, ips, ports):
        with mock.patch("quark.db.api.ip_address_find") as ip_find:
            ip_models = []
            port_models = []
            for port in ports:
                p = models.Port()
                p.update(port)
                port_models.append(p)
                for ip in ips:
                    version = ip.pop("version")
                    ip_mod = models.IPAddress()
                    ip_mod.update(ip)
                    ip_mod.version = version
                    ip_mod.ports = port_models
                    # Set up Port to IP associations
                    assoc = models.PortIpAssociation()
                    assoc.port = p
                    assoc.port_id = p.id
                    assoc.ip_address = ip_mod
                    assoc.ip_address_id = ip_mod.id
                    ip_mod.associations.append(assoc)
                    ip_models.append(ip_mod)
                ip_find.return_value = ip_models
            yield

    def test_get_ip_addresses(self):
        port = dict(id=100, device_id="foobar")
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4)
        with self._stubs(ips=[ip], ports=[port]):
            res = self.plugin.get_ip_addresses(self.context)
            addr_res = res[0]
            self.assertEqual(ip["id"], addr_res["id"])
            self.assertEqual(ip["subnet_id"], addr_res["subnet_id"])
            self.assertEqual(ip["address_readable"], addr_res["ip_address"])
            self.assertEqual(addr_res["port_ids"][0], port["id"])

    def test_get_ip_addresses_multiple(self):
        port = dict(id=100, device_id="foobar")
        ips = [dict(id=1, address=3232235876, address_readable="192.168.1.100",
                    subnet_id=1, network_id=2, version=4),
               dict(id=2, address=3232235878, address_readable="192.168.1.101",
                    subnet_id=1, network_id=2, version=4)]
        with self._stubs(ips=ips, ports=[port]):
            res = self.plugin.get_ip_addresses(self.context)
            self.assertEqual(len(res), 2)
            for i, addr in enumerate(sorted(res, key=lambda x: x['id'])):
                self.assertEqual(ips[i]["id"], addr["id"])
                self.assertEqual(ips[i]["subnet_id"], addr["subnet_id"])
                self.assertEqual(ips[i]["address_readable"],
                                 addr["ip_address"])


@mock.patch("quark.plugin_modules.ip_addresses"
            ".validate_and_fetch_segment")
@mock.patch("quark.plugin_modules.ip_addresses.ipam_driver")
@mock.patch("quark.plugin_modules.ip_addresses.db_api")
class TestQuarkGetIpAddressPort(test_quark_plugin.TestQuarkPlugin):
    def _alloc_stub(self, ip_model):
        def _alloc_ip(context, addr, *args, **kwargs):
            addr.append(ip_model)
        return _alloc_ip

    def test_get_ip_address_ports(self, mock_dbapi, mock_ipam, *args):
        port = dict(mac_address="AA:BB:CC:DD:EE:FF", network_id=1,
                    tenant_id=self.context.tenant_id, device_id=2,
                    bridge="xenbr0", device_owner='network:dhcp', id=100)
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4)
        port_model = models.Port()
        port_model.update(port)
        ip_model = models.IPAddress()
        ip_model.update(ip)

        mock_dbapi.port_find.return_value = [port_model]
        mock_ipam.allocate_ip_address.side_effect = (
            self._alloc_stub(ip_model))

        res = self.plugin.get_ports_for_ip_address(self.context, 1)[0]
        mock_dbapi.port_find.assert_called_with(self.context, None, None, None,
                                                join_security_groups=True,
                                                fields=None,
                                                ip_address_id=[1])

        self.assertEqual(port["id"], res["id"])
        self.assertEqual(port["device_id"], res["device_id"])
        self.assertFalse('mac_address' in res)
        self.assertFalse('network_id' in res)
        self.assertFalse('bridge' in res)
        self.assertFalse('tenant_id' in res)

    def test_get_ip_address_port(self, mock_dbapi, mock_ipam, *args):
        port = dict(mac_address="AA:BB:CC:DD:EE:FF", network_id=1,
                    tenant_id=self.context.tenant_id, device_id=2,
                    bridge="xenbr0", device_owner='network:dhcp', id=100)
        port2 = dict(mac_address="AA:BB:CC:DD:EE:FF", network_id=1,
                     tenant_id=self.context.tenant_id, device_id=2,
                     bridge="xenbr0", device_owner='network:dhcp', id=100)
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4)
        port_model = models.Port()
        port_model.update(port)
        port_model2 = models.Port()
        port_model2.update(port2)
        ip_model = models.IPAddress()
        ip_model.update(ip)

        mock_dbapi.port_find.return_value = port_model
        mock_ipam.allocate_ip_address.side_effect = (
            self._alloc_stub(ip_model))
        mock_dbapi.ip_address_find.return_value = ip_model

        res = self.plugin.get_port_for_ip_address(self.context, 1, 100)

        mock_dbapi.ip_address_find.assert_called_with(self.context,
                                                      scope=mock_dbapi.ONE,
                                                      id=1)
        mock_dbapi.port_find.assert_called_with(self.context, fields=None,
                                                id=100, ip_address_id=[1],
                                                scope=mock_dbapi.ONE)
        self.assertEqual(port["id"], res["id"])
        self.assertEqual(port["device_id"], res["device_id"])
        self.assertFalse('mac_address' in res)
        self.assertFalse('network_id' in res)
        self.assertFalse('bridge' in res)
        self.assertFalse('tenant_id' in res)

    def test_deleting_inactive_shared_ip(self, mock_dbapi, mock_ipam, *args):
        port = dict(id=100, network_id=2,
                    backend_key="derp", device_id="y")
        port2 = dict(id=101, network_id=2,
                     backend_key="derp", device_id="x")
        ip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                  subnet_id=1, network_id=2, version=4, address_type="shared")
        port_model = models.Port()
        port_model2 = models.Port()
        port_model.update(port)
        port_model2.update(port2)
        ip_model = models.IPAddress()
        ip_model.update(ip)
        ip_model.ports = [port_model, port_model2]

        mock_dbapi.port_find.return_value = port_model
        mock_dbapi.ip_address_find.return_value = ip_model
        mock_ipam.allocate_ip_address.side_effect = (
            self._alloc_stub(ip_model))
        self.plugin.delete_ip_address(self.context, 1)
        self.assertFalse(mock_dbapi.ip_address_delete.called)
        self.assertTrue(mock_ipam.deallocate_ip_address.called)

    def test_get_ip_address_no_ip_fails(self, mock_dbapi, mock_ipam, *args):
        mock_dbapi.ip_address_find.return_value = []
        with self.assertRaises(q_exc.IpAddressNotFound):
            self.plugin.get_port_for_ip_address(self.context, 123, 100)

    def test_get_ip_address_no_port_fails(self, mock_dbapi, mock_ipam, *args):
        mock_dbapi.port_find.return_value = []
        with self.assertRaises(n_exc.PortNotFound):
            self.plugin.get_port_for_ip_address(self.context, 123, 100)



from sqlalchemy.orm import exc

from quark.db import api as db_api
from quark.tests.functional.mysql.base import MySqlBaseFunctionalTest


class TestSubnetsAllocationPoolCache(MySqlBaseFunctionalTest):
    def setUp(self):
        super(TestSubnetsAllocationPoolCache, self).setUp()

    def test_subnet_update_set_alloc_pool_cache_concurrency(self):
        subnet = {"cidr": "192.168.10.0/24"}
        subnet_db = db_api.subnet_create(self.context, **subnet)
        self.context.session.flush()

        # establish second session
        old_session = self.context.session
        self.context._session = None

        subnet_to_delete = db_api.subnet_find(
            self.context, id=subnet_db.id, scope=db_api.ONE)
        db_api.subnet_delete(self.context, subnet_to_delete)
        self.context.session.flush()

        # restore first session
        self.context._session = old_session

        try:
            db_api.subnet_update_set_alloc_pool_cache(
                self.context, subnet_db, {"foo": "bar"})
            self.context.session.flush()
        except exc.StaleDataError as e:
            self.fail("Did not expect StaleDataError exception: {0}".format(e))
        self.assertEqual(subnet_db["_allocation_pool_cache"],
                         "{\"foo\": \"bar\"}")



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Quark Pluggable IPAM
"""

import functools
import itertools
import random
import time
import uuid

import netaddr
from neutron.common import exceptions as n_exc_ext
from neutron_lib import exceptions as n_exc
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_db import exception as db_exception
from oslo_log import log as logging
from oslo_utils import timeutils

from quark import billing
from quark.db import api as db_api
from quark.db import ip_types
from quark.db import models
from quark.drivers import floating_ip_registry as registry
from quark import exceptions as q_exc
from quark import network_strategy
from quark import utils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
STRATEGY = network_strategy.STRATEGY

quark_opts = [
    cfg.IntOpt('v6_allocation_attempts',
               default=10,
               help=_('Number of times to retry generating v6 addresses'
                      ' before failure. Also implicitly controls how many'
                      ' v6 addresses we assign to any port, as the random'
                      ' values generated will be the same every time.')),
    cfg.IntOpt("mac_address_retry_max",
               default=20,
               help=_("Number of times to attempt to allocate a new MAC"
                      " address before giving up.")),
    cfg.IntOpt("ip_address_retry_max",
               default=20,
               help=_("Number of times to attempt to allocate a new IP"
                      " address before giving up.")),
    cfg.BoolOpt("ipam_use_synchronization",
                default=False,
                help=_("Configures whether or not to use the experimental"
                       " semaphore logic around IPAM")),
    cfg.BoolOpt("ipam_select_subnet_v6_locking",
                default=True,
                help=_("Controls whether or not SELECT ... FOR UPDATE is used"
                       " when retrieving v6 subnets explicitly."))
]

CONF.register_opts(quark_opts, "QUARK")

# NOTE(mdietz): equivalent to the following line, but converting
#               v6 addresses in netaddr is very slow.
# netaddr.IPAddress("::0200:0:0:0").value
MAGIC_INT = 144115188075855872


def no_synchronization(*args, **kwargs):
    def wrap(f):
        @functools.wraps(f)
        def inner(*args, **kwargs):
            return f(*args, **kwargs)
        return inner
    return wrap


def named(sema):
    return "%s.%s" % (__name__, sema)


if CONF.QUARK.ipam_use_synchronization:
    synchronized = lockutils.synchronized
else:
    synchronized = no_synchronization


def rfc2462_ip(mac, cidr):
    # NOTE(mdietz): see RFC2462
    int_val = netaddr.IPNetwork(cidr).value
    mac = netaddr.EUI(mac)
    LOG.info("Using RFC2462 method to generate a v6 with MAC %s" % mac)
    int_val += mac.eui64().value
    int_val ^= MAGIC_INT
    return int_val


def rfc3041_ip(port_id, cidr):
    if not port_id:
        random_stuff = uuid.uuid4()
    else:
        random_stuff = uuid.UUID(port_id)
    random.seed(int(random_stuff))
    int_val = netaddr.IPNetwork(cidr).value
    while True:
        rand_bits = random.getrandbits(64)
        LOG.info("Using RFC3041 method to generate a v6 with bits %s" %
                 rand_bits)
        val = int_val + rand_bits
        val ^= MAGIC_INT
        yield val


def ip_address_failure(network_id):
    if STRATEGY.is_provider_network(network_id):
        return q_exc.ProviderNetworkOutOfIps(net_id=network_id)
    else:
        return n_exc.IpAddressGenerationFailure(net_id=network_id)


def generate_v6(mac, port_id, cidr):
    # NOTE(mdietz): RM10879 - if we don't have a MAC, don't panic, defer to
    #               our magic rfc3041_ip method instead. If an IP is created
    #               by the ip_addresses controller, we wouldn't necessarily
    #               have a MAC to base our generator on in that case for
    #               example.
    if mac is not None:
        addr = rfc2462_ip(mac, cidr)
        yield addr

    for addr in rfc3041_ip(port_id, cidr):
        yield addr


def ipam_logged(fx):
    def wrap(self, *args, **kwargs):
        log = QuarkIPAMLog()
        kwargs['ipam_log'] = log
        try:
            return fx(self, *args, **kwargs)
        finally:
            log.end()
    return wrap


class QuarkIPAMLog(object):
    def __init__(self):
        self.entries = {}
        self.success = True

    def make_entry(self, fx_name):
        if fx_name not in self.entries:
            self.entries[fx_name] = []
        entry = QuarkIPAMLogEntry(self, fx_name)
        self.entries[fx_name].append(entry)
        return entry

    def _output(self, status, time_total, fails, successes):
        statistics = ("TIME:%f ATTEMPTS:%d PASS:%d FAIL:%d" %
                      (time_total, fails + successes, successes, fails))
        if not self.success:
            LOG.warning("STATUS:FAILED %s" % statistics)
        else:
            LOG.debug("STATUS:SUCCESS %s" % statistics)

    def end(self):
        total = 0
        fails = 0
        successes = 0
        for fx, entries in self.entries.items():
            for entry in entries:
                total += entry.get_time()
                if entry.success:
                    successes += 1
                else:
                    fails += 1
        self._output(self.success, total, fails, successes)

    def failed(self):
        self.success = False


class QuarkIPAMLogEntry(object):
    def __init__(self, log, name):
        self.name = name
        self.log = log
        self.start_time = time.time()
        self.success = True

    def failed(self):
        self.success = False

    def end(self):
        self.end_time = time.time()

    def get_time(self):
        if not hasattr(self, 'end_time'):
            return 0
        return self.end_time - self.start_time


class QuarkIpam(object):
    @synchronized(named("allocate_mac_address"))
    def allocate_mac_address(self, context, net_id, port_id, reuse_after,
                             mac_address=None,
                             use_forbidden_mac_range=False, **kwargs):
        if mac_address:
            mac_address = netaddr.EUI(mac_address).value

        kwargs.update({"network_id": net_id, "port_id": port_id,
                       "mac_address": mac_address,
                       "use_forbidden_mac_range": use_forbidden_mac_range})
        LOG.info(("Attempting to allocate a new MAC address "
                  "[{0}]").format(utils.pretty_kwargs(**kwargs)))

        for retry in xrange(CONF.QUARK.mac_address_retry_max):
            LOG.info("Attemping to reallocate deallocated MAC (step 1 of 3),"
                     " attempt {0} of {1}".format(
                         retry + 1, CONF.QUARK.mac_address_retry_max))
            try:
                with context.session.begin():
                    transaction = db_api.transaction_create(context)
                update_kwargs = {
                    "deallocated": False,
                    "deallocated_at": None,
                    "transaction_id": transaction.id
                }
                filter_kwargs = {
                    "deallocated": True,
                }
                if mac_address is not None:
                    filter_kwargs["address"] = mac_address
                if reuse_after is not None:
                    filter_kwargs["reuse_after"] = reuse_after
                elevated = context.elevated()
                result = db_api.mac_address_reallocate(
                    elevated, update_kwargs, **filter_kwargs)
                if not result:
                    break

                reallocated_mac = db_api.mac_address_reallocate_find(
                    elevated, transaction.id)
                if reallocated_mac:
                    dealloc = netaddr.EUI(reallocated_mac["address"])
                    LOG.info("Found a suitable deallocated MAC {0}".format(
                        str(dealloc)))
                    LOG.info("MAC assignment for port ID {0} completed "
                             "with address {1}".format(port_id, dealloc))
                    return reallocated_mac
            except Exception:
                LOG.exception("Error in mac reallocate...")
                continue

        LOG.info("Couldn't find a suitable deallocated MAC, attempting "
                 "to create a new one")

        # This could fail if a large chunk of MACs were chosen explicitly,
        # but under concurrent load enough MAC creates should iterate without
        # any given thread exhausting its retry count.
        for retry in xrange(CONF.QUARK.mac_address_retry_max):
            LOG.info("Attemping to find a range to create a new MAC in "
                     "(step 2 of 3), attempt {0} of {1}".format(
                         retry + 1, CONF.QUARK.mac_address_retry_max))
            next_address = None
            with context.session.begin():
                try:
                    fn = db_api.mac_address_range_find_allocation_counts
                    mac_range = \
                        fn(context, address=mac_address,
                           use_forbidden_mac_range=use_forbidden_mac_range)

                    if not mac_range:
                        LOG.info("No MAC ranges could be found given "
                                 "the criteria")
                        break

                    rng, addr_count = mac_range
                    LOG.info("Found a MAC range {0}".format(rng["cidr"]))

                    last = rng["last_address"]
                    first = rng["first_address"]
                    if (last - first + 1) <= addr_count:
                        # Somehow, the range got filled up without us
                        # knowing, so set the next_auto_assign to be -1
                        # so we never try to create new ones
                        # in this range
                        db_api.mac_range_update_set_full(context, rng)
                        LOG.info("MAC range {0} is full".format(rng["cidr"]))
                        continue

                    if mac_address:
                        next_address = mac_address
                    else:
                        next_address = rng["next_auto_assign_mac"]
                        if next_address + 1 > rng["last_address"]:
                            db_api.mac_range_update_set_full(context, rng)
                        else:
                            db_api.mac_range_update_next_auto_assign_mac(
                                context, rng)
                        context.session.refresh(rng)
                except Exception:
                    LOG.exception("Error in updating mac range")
                    continue

            # Based on the above, this should only fail if a MAC was
            # was explicitly chosen at some point. As such, fall through
            # here and get in line for a new MAC address to try
            try:
                mac_readable = str(netaddr.EUI(next_address))
                LOG.info("Attempting to create new MAC {0} "
                         "(step 3 of 3)".format(mac_readable))
                with context.session.begin():
                    address = db_api.mac_address_create(
                        context, address=next_address,
                        mac_address_range_id=rng["id"])
                    LOG.info("MAC assignment for port ID {0} completed with "
                             "address {1}".format(port_id, mac_readable))
                    return address
            except Exception:
                LOG.info("Failed to create new MAC {0}".format(mac_readable))
                LOG.exception("Error in creating mac. MAC possibly duplicate")
                continue

        raise n_exc_ext.MacAddressGenerationFailure(net_id=net_id)

    @synchronized(named("reallocate_ip"))
    def attempt_to_reallocate_ip(self, context, net_id, port_id, reuse_after,
                                 version=None, ip_address=None,
                                 segment_id=None, subnets=None, **kwargs):
        version = version or [4, 6]
        elevated = context.elevated()

        LOG.info("Attempting to reallocate an IP (step 1 of 3) - [{0}]".format(
            utils.pretty_kwargs(network_id=net_id, port_id=port_id,
                                version=version, segment_id=segment_id,
                                subnets=subnets, ip_address=ip_address)))

        if version == 6:
            # Defers to the create case. The reason why is we'd have to look
            # up subnets here to correctly generate the v6. If we split them
            # up into reallocate and create, we'd be looking up the same
            # subnets twice, which is a waste of time.

            # TODO(mdietz): after reviewing this code, this block annoyingly
            #               doesn't trigger in the ANY case, since we end up
            #               using a list of [4, 6]. It works as expected most
            #               of the time, but we can anticipate that isolated
            #               networks will end up using sequential assignment.
            #               Probably want to rework this logic to compensate
            #               at some point. Considering they all come from the
            #               same MAC address pool, nothing bad will happen,
            #               just worth noticing and fixing.
            LOG.info("Identified as v6 case, deferring to IP create path")
            return []

        sub_ids = []
        if subnets:
            sub_ids = subnets
        elif segment_id:
            subnets = db_api.subnet_find(elevated,
                                         network_id=net_id,
                                         segment_id=segment_id)
            sub_ids = [s["id"] for s in subnets]
            if not sub_ids:
                LOG.info("No subnets matching segment_id {0} could be "
                         "found".format(segment_id))
                raise ip_address_failure(net_id)

        ip_kwargs = {
            "network_id": net_id,
            "deallocated": True,
            "version": version,
            "lock_id": None,
        }
        if reuse_after is not None:
            ip_kwargs["reuse_after"] = reuse_after
        if ip_address is not None:
            ip_kwargs["ip_address"] = ip_address
            del ip_kwargs["deallocated"]
        if sub_ids:
            ip_kwargs["subnet_id"] = sub_ids

        ipam_log = kwargs.get('ipam_log', None)

        for retry in xrange(CONF.QUARK.ip_address_retry_max):
            attempt = None
            if ipam_log:
                attempt = ipam_log.make_entry("attempt_to_reallocate_ip")
            LOG.info("Attempt {0} of {1}".format(
                retry + 1, CONF.QUARK.ip_address_retry_max))
            try:
                with context.session.begin():
                    transaction = db_api.transaction_create(context)
                m = models.IPAddress
                update_kwargs = {
                    m.transaction_id: transaction.id,
                    m.address_type: kwargs.get("address_type", ip_types.FIXED),
                    m.deallocated: False,
                    m.deallocated_at: None,
                    m.used_by_tenant_id: context.tenant_id,
                    m.allocated_at: timeutils.utcnow(),
                }
                result = db_api.ip_address_reallocate(
                    elevated, update_kwargs, **ip_kwargs)
                if not result:
                    LOG.info("Couldn't update any reallocatable addresses "
                             "given the criteria")
                    if attempt:
                        attempt.failed()
                    break

                updated_address = db_api.ip_address_reallocate_find(
                    elevated, transaction.id)
                if not updated_address:
                    if attempt:
                        attempt.failed()
                    continue

                LOG.info("Address {0} is reallocated".format(
                    updated_address["address_readable"]))
                return [updated_address]
            except Exception:
                if attempt:
                    attempt.failed()
                LOG.exception("Error in reallocate ip...")
            finally:
                if attempt:
                    attempt.end()
        return []

    def is_strategy_satisfied(self, ip_addresses, allocate_complete=False):
        return ip_addresses

    def _allocate_from_subnet(self, context, net_id, subnet,
                              port_id, reuse_after, ip_address=None, **kwargs):

        LOG.info("Creating a new address in subnet {0} - [{1}]".format(
            subnet["_cidr"], utils.pretty_kwargs(network_id=net_id,
                                                 subnet=subnet,
                                                 port_id=port_id,
                                                 ip_address=ip_address)))

        if subnet and subnet["ip_policy"]:
            ip_policy_cidrs = subnet["ip_policy"].get_cidrs_ip_set()
        else:
            ip_policy_cidrs = netaddr.IPSet([])

        next_ip = ip_address
        if not next_ip:
            if subnet["next_auto_assign_ip"] != -1:
                next_ip = netaddr.IPAddress(subnet["next_auto_assign_ip"] - 1)
            else:
                next_ip = netaddr.IPAddress(subnet["last_ip"])
            if subnet["ip_version"] == 4:
                next_ip = next_ip.ipv4()

        LOG.info("Next IP is {0}".format(str(next_ip)))
        if ip_policy_cidrs and next_ip in ip_policy_cidrs and not ip_address:
            LOG.info("Next IP {0} violates policy".format(str(next_ip)))
            raise q_exc.IPAddressPolicyRetryableFailure(ip_addr=next_ip,
                                                        net_id=net_id)
        try:
            with context.session.begin():
                address = db_api.ip_address_create(
                    context, address=next_ip, subnet_id=subnet["id"],
                    deallocated=0, version=subnet["ip_version"],
                    network_id=net_id,
                    port_id=port_id,
                    address_type=kwargs.get('address_type', ip_types.FIXED))
                address["deallocated"] = 0
                # alexm: instead of notifying billing from here we notify from
                # allocate_ip_address() when it's clear that the IP
                # allocation was successful
        except db_exception.DBDuplicateEntry:
            raise n_exc.IpAddressInUse(ip_address=next_ip, net_id=net_id)
        except db_exception.DBError:
            raise q_exc.IPAddressRetryableFailure(ip_addr=next_ip,
                                                  net_id=net_id)

        return address

    def _allocate_from_v6_subnet(self, context, net_id, subnet,
                                 port_id, reuse_after, ip_address=None,
                                 **kwargs):
        """This attempts to allocate v6 addresses as per RFC2462 and RFC3041.

        To accomodate this, we effectively treat all v6 assignment as a
        first time allocation utilizing the MAC address of the VIF. Because
        we recycle MACs, we will eventually attempt to recreate a previously
        generated v6 address. Instead of failing, we've opted to handle
        reallocating that address in this method.

        This should provide a performance boost over attempting to check
        each and every subnet in the existing reallocate logic, as we'd
        have to iterate over each and every subnet returned
        """

        LOG.info("Attempting to allocate a v6 address - [{0}]".format(
            utils.pretty_kwargs(network_id=net_id, subnet=subnet,
                                port_id=port_id, ip_address=ip_address)))

        if ip_address:
            LOG.info("IP %s explicitly requested, deferring to standard "
                     "allocation" % ip_address)
            return self._allocate_from_subnet(context, net_id=net_id,
                                              subnet=subnet, port_id=port_id,
                                              reuse_after=reuse_after,
                                              ip_address=ip_address, **kwargs)
        else:
            mac = kwargs.get("mac_address")
            if mac:
                mac = kwargs["mac_address"].get("address")

            if subnet and subnet["ip_policy"]:
                ip_policy_cidrs = subnet["ip_policy"].get_cidrs_ip_set()
            else:
                ip_policy_cidrs = netaddr.IPSet([])

            for tries, ip_address in enumerate(
                    generate_v6(mac, port_id, subnet["cidr"])):

                LOG.info("Attempt {0} of {1}".format(
                    tries + 1, CONF.QUARK.v6_allocation_attempts))

                if tries > CONF.QUARK.v6_allocation_attempts - 1:
                    LOG.info("Exceeded v6 allocation attempts, bailing")
                    raise ip_address_failure(net_id)

                ip_address = netaddr.IPAddress(ip_address).ipv6()
                LOG.info("Generated a new v6 address {0}".format(
                    str(ip_address)))

                if (ip_policy_cidrs is not None and
                        ip_address in ip_policy_cidrs):
                    LOG.info("Address {0} excluded by policy".format(
                        str(ip_address)))
                    continue

                try:
                    with context.session.begin():
                        address = db_api.ip_address_create(
                            context, address=ip_address,
                            subnet_id=subnet["id"],
                            version=subnet["ip_version"], network_id=net_id,
                            address_type=kwargs.get('address_type',
                                                    ip_types.FIXED))
                        return address
                except db_exception.DBDuplicateEntry:
                    # This shouldn't ever happen, since we hold a unique MAC
                    # address from the previous IPAM step.
                    LOG.info("{0} exists but was already "
                             "allocated".format(str(ip_address)))
                    LOG.debug("Duplicate entry found when inserting subnet_id"
                              " %s ip_address %s", subnet["id"], ip_address)

    def _allocate_ips_from_subnets(self, context, new_addresses, net_id,
                                   subnets, port_id, reuse_after,
                                   ip_address=None, **kwargs):

        LOG.info("Allocating IP(s) from chosen subnet(s) (step 3 of 3) - "
                 "[{0}]".format(utils.pretty_kwargs(
                     network_id=net_id, port_id=port_id,
                     new_addresses=new_addresses, ip_address=ip_address)))

        subnets = subnets or []
        allocated_ips = [ip.get("address_readable") for ip in new_addresses]
        for subnet in subnets:
            if not subnet:
                continue

            if str(ip_address) in allocated_ips:
                continue

            LOG.info("Attempting to allocate from {0} - {1}".format(
                subnet["id"], subnet["_cidr"]))

            address = None
            if int(subnet["ip_version"]) == 4:
                address = self._allocate_from_subnet(context, net_id,
                                                     subnet, port_id,
                                                     reuse_after,
                                                     ip_address, **kwargs)
            else:
                address = self._allocate_from_v6_subnet(context, net_id,
                                                        subnet, port_id,
                                                        reuse_after,
                                                        ip_address, **kwargs)
            if address:
                LOG.info("Created IP {0}".format(
                    address["address_readable"]))
                new_addresses.append(address)

        return new_addresses

    @ipam_logged
    def allocate_ip_address(self, context, new_addresses, net_id, port_id,
                            reuse_after, segment_id=None, version=None,
                            ip_addresses=None, subnets=None, **kwargs):
        elevated = context.elevated()
        subnets = subnets or []
        ip_addresses = ip_addresses or []

        ipam_log = kwargs.get('ipam_log', None)
        LOG.info("Starting a new IP address(es) allocation. Strategy "
                 "is {0} - [{1}]".format(
                     self.get_name(),
                     utils.pretty_kwargs(network_id=net_id, port_id=port_id,
                                         new_addresses=new_addresses,
                                         ip_addresses=ip_addresses,
                                         subnets=subnets,
                                         segment_id=segment_id,
                                         version=version)))

        def _try_reallocate_ip_address(ipam_log, ip_addr=None):
            new_addresses.extend(self.attempt_to_reallocate_ip(
                context, net_id, port_id, reuse_after, version=version,
                ip_address=ip_addr, segment_id=segment_id, subnets=subnets,
                **kwargs))

        def _try_allocate_ip_address(ipam_log, ip_addr=None, sub=None):
            for retry in xrange(CONF.QUARK.ip_address_retry_max):
                attempt = None
                if ipam_log:
                    attempt = ipam_log.make_entry("_try_allocate_ip_address")
                LOG.info("Allocating new IP attempt {0} of {1}".format(
                    retry + 1, CONF.QUARK.ip_address_retry_max))
                if not sub:
                    subnets = self._choose_available_subnet(
                        elevated, net_id, version, segment_id=segment_id,
                        ip_address=ip_addr, reallocated_ips=new_addresses)
                else:
                    subnets = [self.select_subnet(context, net_id,
                                                  ip_addr, segment_id,
                                                  subnet_ids=[sub])]
                LOG.info("Subnet selection returned {0} viable subnet(s) - "
                         "IDs: {1}".format(len(subnets),
                                           ", ".join([str(s["id"])
                                                      for s in subnets if s])))

                try:
                    self._allocate_ips_from_subnets(context, new_addresses,
                                                    net_id, subnets,
                                                    port_id, reuse_after,
                                                    ip_addr, **kwargs)
                except q_exc.IPAddressRetryableFailure:
                    LOG.exception("Error in allocating IP")
                    if attempt:
                        LOG.debug("ATTEMPT FAILED")
                        attempt.failed()
                    remaining = CONF.QUARK.ip_address_retry_max - retry - 1
                    if remaining > 0:
                        LOG.info("{0} retries remain, retrying...".format(
                            remaining))
                    else:
                        LOG.info("No retries remaing, bailing")
                    continue
                finally:
                    if attempt:
                        attempt.end()

                break

        ip_addresses = [netaddr.IPAddress(ip_address)
                        for ip_address in ip_addresses]

        if ip_addresses:
            for ip_address in ip_addresses:
                _try_reallocate_ip_address(ipam_log, ip_address)
        else:
            _try_reallocate_ip_address(ipam_log)

        if self.is_strategy_satisfied(new_addresses):
            return
        else:
            LOG.info("Reallocated addresses {0} but still need more addresses "
                     "to satisfy strategy {1}. Falling back to creating "
                     "IPs".format(new_addresses, self.get_name()))

        if ip_addresses or subnets:
            for ip_address, subnet in itertools.izip_longest(ip_addresses,
                                                             subnets):
                _try_allocate_ip_address(ipam_log, ip_address, subnet)
        else:
            _try_allocate_ip_address(ipam_log)

        if self.is_strategy_satisfied(new_addresses, allocate_complete=True):
            # Only notify when all went well
            for address in new_addresses:
                billing.notify(context, billing.IP_ADD, address, **kwargs)
            LOG.info("IPAM for port ID {0} completed with addresses "
                     "{1}".format(port_id,
                                  [a["address_readable"]
                                   for a in new_addresses]))
            return
        ipam_log.failed()

        raise ip_address_failure(net_id)

    def deallocate_ip_address(self, context, address, **kwargs):
        if address["version"] == 6:
            db_api.ip_address_delete(context, address)
        else:
            address["deallocated"] = 1
            address["address_type"] = None

        billing.notify(context, billing.IP_DEL, address, send_usage=True,
                       **kwargs)

    def deallocate_ips_by_port(self, context, port=None, **kwargs):
        ips_to_remove = []
        for addr in port["ip_addresses"]:
            if "ip_address" in kwargs:
                ip = kwargs["ip_address"]
                if ip != netaddr.IPAddress(int(addr["address"])):
                    continue

            # Note: only deallocate ip if this is the
            # only port mapped
            ips_to_remove.append(addr)

        port["ip_addresses"] = list(
            set(port["ip_addresses"]) - set(ips_to_remove))

        # NCP-1541: We don't need to track v6 IPs the same way. Also, we can't
        # delete them until we've removed the FK on the assoc record first, so
        # we have to flush the current state of the transaction.
        # NOTE(mdietz): this does increase traffic to the db because we need
        #               to flush, fetch the records again and potentially make
        #               another trip to deallocate each IP, but keeping our
        #               indices smaller probably provides more value than the
        #               cost
        # NOTE(aquillin): For floating IPs associated with the port, we do not
        #                 want to deallocate the IP or disassociate the IP from
        #                 the tenant, instead we will disassociate floating's
        #                 fixed IP address.
        context.session.flush()
        deallocated_ips = []
        flip = None
        for ip in ips_to_remove:
            if ip["address_type"] in (ip_types.FLOATING, ip_types.SCALING):
                flip = ip
            else:
                if len(ip["ports"]) == 0:
                    self.deallocate_ip_address(context, ip)
                    deallocated_ips.append(ip.id)
        if flip:
            if flip.fixed_ips and len(flip.fixed_ips) == 1:
                # This is a FLIP or SCIP that is only associated with one
                # port and fixed_ip, so we can safely just disassociate all
                # and remove the flip from unicorn.
                db_api.floating_ip_disassociate_all_fixed_ips(context, flip)
                # NOTE(blogan): I'm not too happy about having do another
                # flush but some test runs showed inconsistent state based on
                # SQLAlchemy caching.
                context.session.add(flip)
                context.session.flush()
                billing.notify(context, billing.IP_DISASSOC, flip, **kwargs)
                driver = registry.DRIVER_REGISTRY.get_driver()
                driver.remove_floating_ip(flip)
            elif len(flip.fixed_ips) > 1:
                # This is a SCIP and we need to diassociate the one fixed_ip
                # from the SCIP and update unicorn with the remaining
                # ports and fixed_ips
                remaining_fixed_ips = []
                for fix_ip in flip.fixed_ips:
                    if fix_ip.id in deallocated_ips:
                        db_api.floating_ip_disassociate_fixed_ip(
                            context, flip, fix_ip)
                        context.session.add(flip)
                        context.session.flush()
                        billing.notify(context, billing.IP_DISASSOC, flip,
                                       **kwargs)
                    else:
                        remaining_fixed_ips.append(fix_ip)
                port_fixed_ips = {}
                for fix_ip in remaining_fixed_ips:
                    # NOTE(blogan): Since this is the flip's fixed_ips it
                    # should be safe to assume there is only one port
                    # associated with it.
                    remaining_port = fix_ip.ports[0]
                    port_fixed_ips[remaining_port.id] = {
                        'port': remaining_port,
                        'fixed_ip': fix_ip
                    }
                driver = registry.DRIVER_REGISTRY.get_driver()
                driver.update_floating_ip(flip, port_fixed_ips)

    # NCP-1509(roaet):
    # - started using admin_context due to tenant not claiming when realloc
    def deallocate_mac_address(self, context, address, **kwargs):
        admin_context = context.elevated()
        mac = db_api.mac_address_find(admin_context, address=address,
                                      scope=db_api.ONE)
        if not mac:
            raise q_exc.MacAddressNotFound(
                mac_address_id=address,
                readable_mac=netaddr.EUI(address))

        if (mac["mac_address_range"] is None or
                mac["mac_address_range"]["do_not_use"]):
            db_api.mac_address_delete(admin_context, mac)
        else:
            db_api.mac_address_update(admin_context, mac, deallocated=True,
                                      deallocated_at=timeutils.utcnow())

    def _select_subnet(self, context, net_id, ip_address, segment_id,
                       subnet_ids, **filters):
        # NCP-1480: Don't need to lock V6 subnets, since we don't use
        # next_auto_assign_ip for them. We already uniquely identified
        # the V6 we're going to get by generating a MAC in a previous step.
        # Also note that this only works under BOTH or BOTH_REQUIRED. ANY
        # does not pass an ip_version
        lock_subnets = True
        if (not CONF.QUARK.ipam_select_subnet_v6_locking and
                "ip_version" in filters and
                int(filters["ip_version"]) == 6):
            lock_subnets = False

        select_api = db_api.subnet_find_ordered_by_most_full
        # TODO(mdietz): Add configurable, alternate subnet selection here
        subnets = select_api(context, net_id, lock_subnets=lock_subnets,
                             segment_id=segment_id, scope=db_api.ALL,
                             subnet_id=subnet_ids, **filters)

        if not subnets:
            LOG.info("No subnets found given the search criteria!")
            return

        # TODO(mdietz): Making this into an iterator because we want to move
        #               to selecting 1 subnet at a time and paginating rather
        #               than the bulk fetch. Without locks, we need to
        #               minimize looking at stale data to save ourselves
        #               some retries. Getting then 1 at a time will
        #               facilitate this.
        for subnet, ips_in_subnet in subnets:
            yield subnet, ips_in_subnet

    def _should_mark_subnet_full(self, context, subnet, ipnet, ip_address,
                                 ips_in_subnet):
        ip = subnet["next_auto_assign_ip"]
        # NOTE(mdietz): When atomically updated, this probably
        #               doesn't need the lower bounds check but
        #               I'm not comfortable removing it yet.
        if (subnet["ip_version"] == 4 and ip < subnet["first_ip"] or
                ip > subnet["last_ip"]):
            return True

        ip_policy = None
        if not ip_address:
            # Policies don't prevent explicit assignment, so we only
            # need to check if we're allocating a new IP
            ip_policy = subnet.get("ip_policy")

        policy_size = ip_policy["size"] if ip_policy else 0

        if ipnet.size > (ips_in_subnet + policy_size - 1):
            return False
        return True

    def _ip_in_subnet(self, subnet, subnet_ids, ipnet, ip_address):
        if ip_address:
            requested_ip = netaddr.IPAddress(ip_address)
            if ipnet.version == 4 and requested_ip.version != 4:
                requested_ip = requested_ip.ipv4()
            if requested_ip not in ipnet:
                if subnet_ids is not None:
                    LOG.info("Requested IP {0} not in subnet {1}, "
                             "retrying".format(str(requested_ip),
                                               str(ipnet)))
                    raise q_exc.IPAddressNotInSubnet(
                        ip_addr=ip_address, subnet_id=subnet["id"])
                return False
        return True

    def select_subnet(self, context, net_id, ip_address, segment_id,
                      subnet_ids=None, **filters):
        LOG.info("Selecting subnet(s) - (Step 2 of 3) [{0}]".format(
            utils.pretty_kwargs(network_id=net_id, ip_address=ip_address,
                                segment_id=segment_id, subnet_ids=subnet_ids,
                                ip_version=filters.get("ip_version"))))

        # TODO(mdietz): Invert the iterator and the session, should only be
        #               one subnet per attempt. We should also only be fetching
        #               the subnet and usage when we need to. Otherwise
        #               we're locking every subnet for a segment, and once
        #               we stop locking, we're looking at stale data.
        with context.session.begin():
            for subnet, ips_in_subnet in self._select_subnet(context, net_id,
                                                             ip_address,
                                                             segment_id,
                                                             subnet_ids,
                                                             **filters):
                if subnet is None:
                    continue
                ipnet = netaddr.IPNetwork(subnet["cidr"])
                LOG.info("Trying subnet ID: {0} - CIDR: {1}".format(
                    subnet["id"], subnet["_cidr"]))

                if not self._ip_in_subnet(subnet, subnet_ids, ipnet,
                                          ip_address):
                    continue

                if self._should_mark_subnet_full(context, subnet, ipnet,
                                                 ip_address, ips_in_subnet):
                    LOG.info("Marking subnet {0} as full".format(subnet["id"]))
                    updated = db_api.subnet_update_set_full(context, subnet)

                    # Ensure the session is aware of the changes to the subnet
                    if updated:
                        context.session.refresh(subnet)
                    continue

                if not ip_address and subnet["ip_version"] == 4:
                    auto_inc = db_api.subnet_update_next_auto_assign_ip
                    updated = auto_inc(context, subnet)

                    if updated:
                        context.session.refresh(subnet)
                    else:
                        # This means the subnet was marked full
                        # while we were checking out policies.
                        # Fall out and go back to the outer retry
                        # loop.
                        return

                LOG.info("Subnet {0} - {1} {2} looks viable, "
                         "returning".format(subnet["id"], subnet["_cidr"],
                                            subnet["next_auto_assign_ip"]))
                return subnet


class QuarkIpamANY(QuarkIpam):
    @classmethod
    def get_name(self):
        return "ANY"

    def _choose_available_subnet(self, context, net_id, version=None,
                                 segment_id=None, ip_address=None,
                                 reallocated_ips=None):
        filters = {}
        if version:
            filters["ip_version"] = version
        subnet = self.select_subnet(context, net_id, ip_address, segment_id,
                                    **filters)
        if subnet:
            return [subnet]
        raise ip_address_failure(net_id)


class QuarkIpamBOTH(QuarkIpam):
    @classmethod
    def get_name(self):
        return "BOTH"

    def is_strategy_satisfied(self, reallocated_ips, allocate_complete=False):
        req = [4, 6]
        for ip in reallocated_ips:
            if ip is not None:
                req.remove(ip["version"])
        ips_allocated = len(req)
        if ips_allocated == 0:
            return True
        elif ips_allocated == 1 and allocate_complete:
            return True

        return False

    def attempt_to_reallocate_ip(self, context, net_id, port_id,
                                 reuse_after, version=None,
                                 ip_address=None, segment_id=None,
                                 subnets=None, **kwargs):
        ip_address_version = 4 if not ip_address else ip_address.version
        # NOTE(quade): We do not attempt to reallocate ipv6, so just return
        if ip_address_version == 6:
            return []
        return super(QuarkIpamBOTH, self).attempt_to_reallocate_ip(
            context, net_id, port_id, reuse_after, ip_address_version,
            ip_address, segment_id, subnets=subnets, **kwargs)

    def _choose_available_subnet(self, context, net_id, version=None,
                                 segment_id=None, ip_address=None,
                                 reallocated_ips=None):
        both_subnet_versions = []
        need_versions = [4, 6]
        for i in reallocated_ips:
            if i["version"] in need_versions:
                need_versions.remove(i["version"])
        filters = {}
        for ver in need_versions:
            filters["ip_version"] = ver
            sub = self.select_subnet(context, net_id, ip_address, segment_id,
                                     **filters)
            if sub:
                both_subnet_versions.append(sub)
        if not reallocated_ips and not both_subnet_versions:
            raise ip_address_failure(net_id)

        return both_subnet_versions


class QuarkIpamBOTHREQ(QuarkIpamBOTH):
    @classmethod
    def get_name(self):
        return "BOTH_REQUIRED"

    def is_strategy_satisfied(self, reallocated_ips, allocate_complete=False):
        req = [4, 6]
        for ip in reallocated_ips:
            if ip is not None:
                req.remove(ip["version"])
            ips_allocated = len(req)
            if ips_allocated == 0:
                return True

        return False

    def _choose_available_subnet(self, context, net_id, version=None,
                                 segment_id=None, ip_address=None,
                                 reallocated_ips=None):
        subnets = super(QuarkIpamBOTHREQ, self)._choose_available_subnet(
            context, net_id, version, segment_id, ip_address, reallocated_ips)

        if len(reallocated_ips) + len(subnets) < 2:
            raise ip_address_failure(net_id)
        return subnets


class IronicIpam(QuarkIpam):
    """IPAM base class for the Ironic driver.

    The idea here is that there are many small subnets created for a
    particular segment for a provider network. The Ironic IPAM
    family selects unused ones, and only allows a single allocation
    per subnet.
    """
    def _select_subnet(self, context, net_id, ip_address, segment_id,
                       subnet_ids, **filters):

        lock_subnets = True

        select_api = db_api.subnet_find_unused
        subnets = select_api(context, net_id, lock_subnets=lock_subnets,
                             segment_id=segment_id, scope=db_api.ALL,
                             subnet_id=subnet_ids, **filters)

        if not subnets:
            LOG.info("No subnets found given the search criteria!")
            return

        for subnet, ips_in_subnet in subnets:
            # make sure we don't select subnets that have allocated ips.
            if ips_in_subnet:
                continue
            yield subnet, ips_in_subnet


class IronicIpamANY(IronicIpam, QuarkIpamANY):
    @classmethod
    def get_name(self):
        return "IRONIC_ANY"


class IronicIpamBOTH(IronicIpam, QuarkIpamBOTH):
    @classmethod
    def get_name(self):
        return "IRONIC_BOTH"


class IronicIpamBOTHREQ(IronicIpam, QuarkIpamBOTHREQ):
    @classmethod
    def get_name(self):
        return "IRONIC_BOTH_REQUIRED"


class IpamRegistry(object):
    def __init__(self):
        self.strategies = {
            QuarkIpamANY.get_name(): QuarkIpamANY(),
            QuarkIpamBOTH.get_name(): QuarkIpamBOTH(),
            QuarkIpamBOTHREQ.get_name(): QuarkIpamBOTHREQ(),
            IronicIpamANY.get_name(): IronicIpamANY(),
            IronicIpamBOTH.get_name(): IronicIpamBOTH(),
            IronicIpamBOTHREQ.get_name(): IronicIpamBOTHREQ()
        }

    def is_valid_strategy(self, strategy_name):
        if strategy_name in self.strategies:
            return True
        return False

    def get_strategy(self, strategy_name):
        if self.is_valid_strategy(strategy_name):
            return self.strategies[strategy_name]
        fallback = CONF.QUARK.default_ipam_strategy
        LOG.warn("IPAM strategy %s not found, "
                 "using default %s" % (strategy_name, fallback))
        return self.strategies[fallback]


IPAM_REGISTRY = IpamRegistry()



import json
import mock
import netaddr
from neutron_lib import exceptions as n_exc

from quark import exceptions as qexceptions
from quark.tests.functional.plugin_modules import test_floating_ips


class TestScalingIP(test_floating_ips.BaseFloatingIPTest):

    def setUp(self):
        super(TestScalingIP, self).setUp()
        self.scaling_network = self.floating_network

    def test_create_scaling_ip(self):
        scaling_ip = dict(
            scaling_network_id=self.scaling_network.id,
            ports=[dict(port_id=self.user_port1['id']),
                   dict(port_id=self.user_port2['id'])]
        )
        scaling_ip = {'scalingip': scaling_ip}
        scip = self.plugin.create_scalingip(self.context, scaling_ip)
        self.assertIn(netaddr.IPAddress(scip['scaling_ip_address']),
                      list(netaddr.IPNetwork(self.pub_net_cidr)))
        self.assertEqual(self.scaling_network['id'],
                         scip['scaling_network_id'])
        self.assertEqual(2, len(scip['ports']))
        scip_ports = {scip_port['port_id']: scip_port['fixed_ip_address']
                      for scip_port in scip['ports']}
        port1_fixed_ip = self.user_port1['fixed_ips'][0]['ip_address']
        port2_fixed_ip = self.user_port2['fixed_ips'][0]['ip_address']
        self.assertIn(self.user_port1['id'], scip_ports)
        self.assertIn(self.user_port2['id'], scip_ports)
        self.assertIn(port1_fixed_ip, scip_ports.values())
        self.assertIn(port2_fixed_ip, scip_ports.values())
        self.mock_requests.post.assert_called_once_with(
            self.FAKE_UNICORN_URL, data=mock.ANY, timeout=2
        )
        actual_body = json.loads(self.mock_requests.post.call_args[1]['data'])
        unicorn_body = self._build_expected_unicorn_request_body(
            scip['scaling_ip_address'], [self.user_port1, self.user_port2],
            actual_body=actual_body
        )
        self.assertEqual(unicorn_body, actual_body,
                         msg="Request to the unicorn API is not what is "
                             "expected.")
        get_scip = self.plugin.get_scalingip(self.context, scip['id'])
        self.assertEqual(scip['scaling_ip_address'],
                         get_scip['scaling_ip_address'])

    def test_create_with_invalid_scaling_network_id(self):
        scaling_ip = dict(
            scaling_network_id='some-wrong-network-id',
            ports=[dict(port_id=self.user_port1['id']),
                   dict(port_id=self.user_port2['id'])]
        )
        self.assertRaises(n_exc.NetworkNotFound,
                          self.plugin.create_scalingip,
                          self.context, {"scalingip": scaling_ip})

    def test_create_with_scaling_network_invalid_segment(self):
        scaling_ip = dict(
            scaling_network_id=self.user_network['id'],
            ports=[dict(port_id=self.user_port1['id']),
                   dict(port_id=self.user_port2['id'])]
        )
        self.assertRaises(n_exc.IpAddressGenerationFailure,
                          self.plugin.create_scalingip,
                          self.context, {"scalingip": scaling_ip})

    def test_update_scaling_ip_add_port(self):
        scaling_ip = dict(
            scaling_network_id=self.scaling_network.id,
            ports=[dict(port_id=self.user_port1['id'])]
        )
        scaling_ip = {'scalingip': scaling_ip}
        scip = self.plugin.create_scalingip(self.context, scaling_ip)
        self.mock_requests.reset_mock()
        scaling_ip = dict(ports=[dict(port_id=self.user_port1['id']),
                                 dict(port_id=self.user_port2['id'])])
        updated_scip = self.plugin.update_scalingip(
            self.context, scip['id'], {"scalingip": scaling_ip})
        self.assertEqual(self.scaling_network['id'],
                         updated_scip['scaling_network_id'])
        self.assertEqual(updated_scip['scaling_ip_address'],
                         scip['scaling_ip_address'])
        self.assertEqual(2, len(updated_scip['ports']))
        scip_ports = {scip_port['port_id']: scip_port['fixed_ip_address']
                      for scip_port in updated_scip['ports']}
        port1_fixed_ip = self.user_port1['fixed_ips'][0]['ip_address']
        port2_fixed_ip = self.user_port2['fixed_ips'][0]['ip_address']
        self.assertIn(self.user_port1['id'], scip_ports)
        self.assertIn(self.user_port2['id'], scip_ports)
        self.assertIn(port1_fixed_ip, scip_ports.values())
        self.assertIn(port2_fixed_ip, scip_ports.values())
        self.assertFalse(self.mock_requests.post.called)
        self.assertFalse(self.mock_requests.delete.called)
        expected_url = '/'.join([self.FAKE_UNICORN_URL,
                                 scip['scaling_ip_address']])
        self.mock_requests.put.assert_called_once_with(
            expected_url, data=mock.ANY, timeout=2)
        actual_body = json.loads(self.mock_requests.put.call_args[1]['data'])
        unicorn_body = self._build_expected_unicorn_request_body(
            scip['scaling_ip_address'], [self.user_port1, self.user_port2],
            actual_body=actual_body
        )
        self.assertEqual(unicorn_body, actual_body,
                         msg="Request to the unicorn API is not what is "
                             "expected.")

    def test_update_scaling_ip_remove_port_with_remaining_ports(self):
        scaling_ip = dict(
            scaling_network_id=self.scaling_network.id,
            ports=[dict(port_id=self.user_port1['id']),
                   dict(port_id=self.user_port2['id'])]
        )
        scaling_ip = {'scalingip': scaling_ip}
        scip = self.plugin.create_scalingip(self.context, scaling_ip)
        self.mock_requests.reset_mock()
        scaling_ip = dict(ports=[dict(port_id=self.user_port1['id'])])
        updated_scip = self.plugin.update_scalingip(
            self.context, scip['id'], {"scalingip": scaling_ip})
        self.assertEqual(self.scaling_network['id'],
                         updated_scip['scaling_network_id'])
        self.assertEqual(updated_scip['scaling_ip_address'],
                         scip['scaling_ip_address'])
        self.assertEqual(1, len(updated_scip['ports']))
        scip_ports = {scip_port['port_id']: scip_port['fixed_ip_address']
                      for scip_port in updated_scip['ports']}
        port1_fixed_ip = self.user_port1['fixed_ips'][0]['ip_address']
        self.assertIn(self.user_port1['id'], scip_ports)
        self.assertIn(port1_fixed_ip, scip_ports.values())
        expected_url = '/'.join([self.FAKE_UNICORN_URL,
                                 scip['scaling_ip_address']])
        self.assertFalse(self.mock_requests.post.called)
        self.assertFalse(self.mock_requests.delete.called)
        self.mock_requests.put.assert_called_once_with(
            expected_url, data=mock.ANY, timeout=2)
        actual_body = json.loads(self.mock_requests.put.call_args[1]['data'])
        unicorn_body = self._build_expected_unicorn_request_body(
            scip['scaling_ip_address'], [self.user_port1],
            actual_body=actual_body
        )
        self.assertEqual(unicorn_body, actual_body,
                         msg="Request to the unicorn API is not what is "
                             "expected.")

    def test_update_scaling_ip_clear_ports(self):
        scaling_ip = dict(
            scaling_network_id=self.scaling_network.id,
            ports=[dict(port_id=self.user_port1['id']),
                   dict(port_id=self.user_port2['id'])]
        )
        scaling_ip = {'scalingip': scaling_ip}
        scip = self.plugin.create_scalingip(self.context, scaling_ip)
        self.mock_requests.reset_mock()
        scaling_ip = dict(ports=[])
        updated_scip = self.plugin.update_scalingip(
            self.context, scip['id'], {"scalingip": scaling_ip})
        self.assertEqual(self.scaling_network['id'],
                         updated_scip['scaling_network_id'])
        self.assertEqual(updated_scip['scaling_ip_address'],
                         scip['scaling_ip_address'])
        self.assertEqual(0, len(updated_scip['ports']))
        expected_url = '/'.join([self.FAKE_UNICORN_URL,
                                 scip['scaling_ip_address']])
        self.assertFalse(self.mock_requests.post.called)
        self.assertFalse(self.mock_requests.put.called)
        self.mock_requests.delete.assert_called_once_with(
            expected_url, timeout=2)

    def test_update_scaling_ip_add_ports_from_none(self):
        scaling_ip = dict(
            scaling_network_id=self.scaling_network.id,
            ports=[]
        )
        scaling_ip = {'scalingip': scaling_ip}
        scip = self.plugin.create_scalingip(self.context, scaling_ip)
        self.mock_requests.reset_mock()
        scaling_ip = dict(ports=[dict(port_id=self.user_port1['id']),
                                 dict(port_id=self.user_port2['id'])])
        updated_scip = self.plugin.update_scalingip(
            self.context, scip['id'], {"scalingip": scaling_ip})
        self.assertEqual(self.scaling_network['id'],
                         updated_scip['scaling_network_id'])
        self.assertEqual(updated_scip['scaling_ip_address'],
                         scip['scaling_ip_address'])
        self.assertEqual(2, len(updated_scip['ports']))
        scip_ports = {scip_port['port_id']: scip_port['fixed_ip_address']
                      for scip_port in updated_scip['ports']}
        port1_fixed_ip = self.user_port1['fixed_ips'][0]['ip_address']
        port2_fixed_ip = self.user_port2['fixed_ips'][0]['ip_address']
        self.assertIn(self.user_port1['id'], scip_ports)
        self.assertIn(self.user_port2['id'], scip_ports)
        self.assertIn(port1_fixed_ip, scip_ports.values())
        self.assertIn(port2_fixed_ip, scip_ports.values())
        self.assertFalse(self.mock_requests.put.called)
        self.assertFalse(self.mock_requests.delete.called)
        self.mock_requests.post.assert_called_once_with(
            self.FAKE_UNICORN_URL, data=mock.ANY, timeout=2)
        actual_body = json.loads(self.mock_requests.post.call_args[1]['data'])
        unicorn_body = self._build_expected_unicorn_request_body(
            scip['scaling_ip_address'], [self.user_port1, self.user_port2],
            actual_body=actual_body
        )
        self.assertEqual(unicorn_body, actual_body,
                         msg="Request to the unicorn API is not what is "
                             "expected.")

    @mock.patch('quark.billing.notify')
    @mock.patch('quark.billing.build_payload', return_value={})
    def test_delete_scaling_ip(self, notify, build_payload):
        scaling_ip = dict(
            scaling_network_id=self.scaling_network.id,
            ports=[dict(port_id=self.user_port1['id']),
                   dict(port_id=self.user_port2['id'])]
        )
        scip = self.plugin.create_scalingip(
            self.context, {"scalingip": scaling_ip})
        self.plugin.delete_scalingip(self.context, scip['id'])
        expected_url = '/'.join([self.FAKE_UNICORN_URL,
                                 scip['scaling_ip_address']])
        self.mock_requests.delete.assert_called_once_with(
            expected_url, timeout=2
        )
        self.assertRaises(qexceptions.ScalingIpNotFound,
                          self.plugin.get_scalingip, self.context, scip['id'])
        scips = self.plugin.get_scalingips(self.context)
        self.assertEqual(0, len(scips))

    def test_scaling_ip_not_in_floating_ip_list(self):
        scaling_ip = dict(
            scaling_network_id=self.scaling_network.id,
            ports=[dict(port_id=self.user_port1['id'])]
        )
        scaling_ip = {'scalingip': scaling_ip}
        self.plugin.create_scalingip(self.context, scaling_ip)
        flips = self.plugin.get_floatingips(self.context)
        self.assertEqual(0, len(flips))

    def test_floating_ip_not_in_scaling_ip_list(self):
        floating_ip = dict(
            floating_network_id=self.scaling_network.id,
            port_id=self.user_port1['id']
        )
        floating_ip = {'floatingip': floating_ip}
        self.plugin.create_floatingip(self.context, floating_ip)
        scips = self.plugin.get_scalingips(self.context)
        self.assertEqual(0, len(scips))

    def test_delete_port_associated_with_scip_with_multiple_ports(self):
        scaling_ip = dict(
            scaling_network_id=self.scaling_network.id,
            ports=[dict(port_id=self.user_port1['id']),
                   dict(port_id=self.user_port2['id'])]
        )
        scip = self.plugin.create_scalingip(
            self.context, {"scalingip": scaling_ip})
        self.mock_requests.reset_mock()
        self.context.session.expire_all()
        self.plugin.delete_port(self.context, self.user_port1['id'])
        after_scip = self.plugin.get_scalingip(self.context, scip['id'])
        self.assertEqual(1, len(after_scip['ports']))
        self.assertEqual(self.user_port2['id'],
                         after_scip['ports'][0]['port_id'])
        expected_url = '/'.join([self.FAKE_UNICORN_URL,
                                 scip['scaling_ip_address']])
        self.assertFalse(self.mock_requests.post.called)
        self.assertFalse(self.mock_requests.delete.called)
        self.mock_requests.put.assert_called_once_with(
            expected_url, data=mock.ANY, timeout=2)
        actual_body = json.loads(self.mock_requests.put.call_args[1]['data'])
        unicorn_body = self._build_expected_unicorn_request_body(
            scip['scaling_ip_address'], [self.user_port2],
            actual_body=actual_body
        )
        self.assertEqual(unicorn_body, actual_body,
                         msg="Request to the unicorn API is not what is "
                             "expected.")

    def test_delete_port_associated_with_scip_with_one_port(self):
        scaling_ip = dict(
            scaling_network_id=self.scaling_network.id,
            ports=[dict(port_id=self.user_port1['id'])]
        )
        scip = self.plugin.create_scalingip(
            self.context, {"scalingip": scaling_ip})
        self.mock_requests.reset_mock()
        self.context.session.expire_all()
        self.plugin.delete_port(self.context, self.user_port1['id'])
        after_scip = self.plugin.get_scalingip(self.context, scip['id'])
        self.assertEqual(0, len(after_scip['ports']))
        expected_url = '/'.join([self.FAKE_UNICORN_URL,
                                 scip['scaling_ip_address']])
        self.assertFalse(self.mock_requests.post.called)
        self.assertFalse(self.mock_requests.put.called)
        self.mock_requests.delete.assert_called_once_with(
            expected_url, timeout=2)



"""Add first, last IP to IPPolicyCIDR

Revision ID: 1acd075bd7e1
Revises: 28e55acaf366
Create Date: 2014-08-07 06:52:39.575301

"""

# revision identifiers, used by Alembic.
revision = '1acd075bd7e1'
down_revision = '28e55acaf366'

from alembic import op
import sqlalchemy as sa

from quark.db.custom_types import INET


def upgrade():
    op.add_column('quark_ip_policy_cidrs',
                  sa.Column('first_ip', INET(), nullable=True))
    op.add_column('quark_ip_policy_cidrs',
                  sa.Column('last_ip', INET(), nullable=True))


def downgrade():
    op.drop_column('quark_ip_policy_cidrs', 'last_ip')
    op.drop_column('quark_ip_policy_cidrs', 'first_ip')
